#!/usr/bin/env python3
"""
Utility functions for JensenLab REST APIs.
https://api.jensenlab.org/Textmining?type1=-26&id1=DOID:10652&type2=9606&limit=10&format=json
https://api.jensenlab.org/Textmining?query=jetlag[tiab]%20OR%20jet-lag[tiab]&type2=9606&limit=10&format=json
"""
import sys,os,re,json,time,logging

from ..util import rest_utils
#
##############################################################################
def GetDiseaseGenes(base_url, ids, fout):
  n_out=0; tags=None;
  for id_this in ids:
    rval = rest_utils.GetURL(base_url+'/Textmining?type1=-26&id1={0}&type2=9606&limit=10&format=json'.format(id_this), parse_json=True)
    genes = rval[0] #dict
    ensgs = list(genes.keys())
    flag = rval[1] #?
    for ensg in ensgs:
      gene = genes[ensg]
      logging.debug(json.dumps(gene, indent=2))
      if not tags:
        tags = list(gene.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(gene[tag]) if tag in gene else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
  logging.info("n_out: %d"%(n_out))

##############################################################################
def GetPubmedComentionGenes(base_url, ids, fout):
  """Search by co-mentioned terms."""
  n_out=0; tags=None;
  for id_this in ids:
    rval = rest_utils.GetURL(base_url+'/Textmining?query={0}[tiab]&type2=9606&limit=10&format=json'.format(id_this), parse_json=True)
    genes = rval[0] #dict
    ensgs = list(genes.keys())
    flag = rval[1] #?
    for ensg in ensgs:
      gene = genes[ensg]
      logging.debug(json.dumps(gene, indent=2))
      if not tags:
        tags = list(gene.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [(str(gene[tag]) if tag in gene else "") for tag in tags]
      fout.write("\t".join(vals)+"\n")
  logging.info("n_out: %d"%(n_out))

##############################################################################



#!/usr/bin/env python3
'''
Utility functions for original SOAP (pre-REST) PUG API.

Jeremy Yang
'''
import sys,os,re,urllib,time,logging
import xml.dom.minidom

from ...util import xml_utils

PUGURL='http://pubchem.ncbi.nlm.nih.gov/pug/pug.cgi'

#############################################################################
def PollPug(reqid,mode='status',pugurl=PUGURL,ntries=20,poll_wait=10):
  qxml='''\
<PCT-Data><PCT-Data_input>
  <PCT-InputData><PCT-InputData_request>
    <PCT-Request>
      <PCT-Request_reqid>%(REQID)d</PCT-Request_reqid>
      <PCT-Request_type value="%(MODE)s"/>
    </PCT-Request>
  </PCT-InputData_request></PCT-InputData>
</PCT-Data_input></PCT-Data>
'''%{'REQID':reqid,'MODE':mode}
  logging.debug('connecting %s...'%pugurl)
  f=UrlOpenTry(pugurl,qxml,ntries,poll_wait)
  pugxml=f.read()
  f.close()
  status,reqid,url,qkey,wenv,error = ParsePugXml(pugxml)

  return status,reqid,url,qkey,wenv,error

#############################################################################
def ParsePugXml(pugxml):
  status,reqid,url,qkey,wenv,error=None,None,None,None,None,None
  try:
    dom=xml.dom.minidom.parseString(pugxml)
  except xml.parsers.expat.ExpatError as e:
    logging.info('XML parse error: %s'%e)
    logging.info('XML: %s'%pugxml)
    sys.exit(1)
  tag='PCT-Status'
  statuss=xml_utils.DOM_GetAttr(dom,tag,'value')
  if statuss: status=statuss[0]
  tag='PCT-Status-Message_message'
  errors=xml_utils.DOM_GetLeafValsByTagName(dom,tag)
  if errors: error=errors[0]
  tag='PCT-Waiting_reqid'
  reqids=xml_utils.DOM_GetLeafValsByTagName(dom,tag)
  if reqids: reqid=int(reqids[0])
  tag='PCT-Download-URL_url'
  urls=xml_utils.DOM_GetLeafValsByTagName(dom,tag)
  if urls: url=urls[0]
  tag='PCT-Entrez_query-key'
  qkeys=xml_utils.DOM_GetLeafValsByTagName(dom,tag)
  if qkeys: qkey=qkeys[0]
  tag='PCT-Entrez_webenv'
  wenvs=xml_utils.DOM_GetLeafValsByTagName(dom,tag)
  if wenvs: wenv=wenvs[0]
  return status,reqid,url,qkey,wenv,error

#############################################################################
def UrlOpenTry(url=PUGURL,xml=None,ntries=20,poll_wait=10):
  ### To recover from FTP timeouts:
  fin=None
  for i in range(ntries):
    try:
      if xml:
        fin=urllib.urlopen(url,xml)
      else:
        fin=urllib.urlopen(url)
      break
    except IOError as e:
      logging.info('IOError: %s'%e)
      time.sleep(poll_wait)
  if not fin:
    logging.info('ERROR: failed download, %d tries, quitting...'%ntries)
    sys.exit(1)
  return fin

#############################################################################
def DownloadUrl(url,fout,ntries=20,poll_wait=10):
  fin=UrlOpenTry(url,None,ntries,poll_wait)
  nbytes=0
  while True:
    buff=fin.read(1024)
    if not buff: break
    fout.write(buff)
    nbytes+=len(buff)
  fin.close()
  return nbytes

#############################################################################
### error status values include: "server-error"
#############################################################################
def CheckStatus(status,error):
  if status not in ('success','queued','running'):
    ErrorExit('query failed; quitting (status="%s",error="%s").'%(status,error))

#############################################################################
def ErrorExit(msg):
  logging.info(msg)
  sys.exit(1)



#!/usr/bin/env python3
"""
OpenTargets REST API client, using the Python client package.

https://www.targetvalidation.org/documentation/api/
http://opentargets.readthedocs.io/
https://docs.targetvalidation.org/programmatic-access/python-client
https://docs.targetvalidation.org/programmatic-access/rest-api
https://docs.targetvalidation.org/data-sources/genetic-associations
https://platform-api.opentargets.io/v3/platform
https://platform-api.opentargets.io/v3/platform/docs/swagger-ui
"""
import sys,os,re,time,argparse,json,csv,logging

import opentargets

#############################################################################
def SearchAssociations(otclient, ids, idtype, minscore, skip, nmax, fout):
  csvWriter = csv.writer(fout, delimiter='\t', quotechar='"', quoting=csv.QUOTE_MINIMAL)
  tags=None;
  n_id=0; n_assn=0; n_id_found=0;
  for id_this in ids:
    n_id+=1
    if n_id<=skip: continue
    logging.debug('%s:'%id_this)
    try:
      if idtype=="disease":
        assns = otclient.get_associations_for_disease(id_this)
      else:
        assns = otclient.get_associations_for_target(id_this)
    except Exception as e:
      logging.error(str(e))
      continue
    if assns:
      n_id_found+=1
    for assn in assns:
      logging.debug(json.dumps(assn, sort_keys=True, indent=2))
      if not tags:
        tags=["id","is_direct","target_id","gene_name","gene_symbol","disease_id","disease_efo_label","score_overall"]
        tags_assn_score_sources = list(assn['association_score']['datasources'].keys())
        tags_assn_score_types = list(assn['association_score']['datatypes'].keys())
        tags_evidence_count_sources = list(assn['evidence_count']['datasources'].keys())
        tags_evidence_count_types = list(assn['evidence_count']['datatypes'].keys())
        csvWriter.writerow(tags+
		list(map(lambda t:'assn_score_source.%s'%t, tags_assn_score_sources))+
		list(map(lambda t:'assn_score_type.%s'%t, tags_assn_score_types))+
		list(map(lambda t:'evidence_count_source.%s'%t, tags_evidence_count_sources))+
		list(map(lambda t:'evidence_count_type.%s'%t, tags_evidence_count_types)))
      assn_id = assn['id'] if 'id' in assn else ''
      is_direct = assn['is_direct'] if 'is_direct' in assn else ''
      target_id = assn['target']['id'] if 'target' in assn and 'id' in assn['target'] else ''
      gene_name = assn['target']['gene_info']['name'] if 'target' in assn and 'gene_info' in assn['target'] and 'name' in assn['target']['gene_info'] else ''
      gene_symbol = assn['target']['gene_info']['symbol'] if 'target' in assn and 'gene_info' in assn['target'] and 'symbol' in assn['target']['gene_info'] else ''
      disease_id = assn['disease']['id'] if 'disease' in assn and 'id' in assn['disease'] else ''
      disease_efo_label = assn['disease']['efo_info']['label'] if 'disease' in assn and 'efo_info' in assn['disease'] and 'label' in assn['disease']['efo_info'] else ''
      score_overall = assn['association_score']['overall'] if 'association_score' in assn and 'overall' in assn['association_score'] else ''
      vals=[assn_id, is_direct, target_id, gene_name, gene_symbol, disease_id, disease_efo_label, score_overall]
      for tag in tags_assn_score_sources:
        vals.append(assn['association_score']['datasources'][tag] if tag in assn['association_score']['datasources'] else '')
      for tag in tags_assn_score_types:
        vals.append(assn['association_score']['datatypes'][tag] if tag in assn['association_score']['datatypes'] else '')
      for tag in tags_evidence_count_sources:
        vals.append(assn['evidence_count']['datasources'][tag] if tag in assn['evidence_count']['datasources'] else '')
      for tag in tags_evidence_count_types:
        vals.append(assn['evidence_count']['datatypes'][tag] if tag in assn['evidence_count']['datatypes'] else '')

      for i in range(len(vals)):
        if type(vals[i]) is float:
          vals[i] = round(vals[i], 5)
      csvWriter.writerow(vals)
      n_assn+=1
    logging.debug('n_id=%d ; skip=%d ; nmax=%d'%(n_id, skip, nmax))
    if nmax and n_id-skip>=nmax: break
  logging.info('n_id = %d; n_id_found = %d; n_assn = %d'%(n_id-skip, n_id_found, n_assn))

#############################################################################
if __name__=='__main__':
  IDTYPES = ["gene", "disease"]
  #SOURCES = ["cancer_gene_census", "chembl", "crispr", "europepmc", "eva", "eva_somatic", "expression_atlas", "gene2phenotype", "genomics_england", "gwas_catalog", "intogen", "ot_genetics_portal", "phenodigm", "phewas_catalog", "postgap", "progeny", "reactome", "slapenrich", "sysbio", "uniprot", "uniprot_literature", "uniprot_somatic"]
  epilog = """Examples:
Gene IDs: ENSG00000163914, ENSG00000072110;
Disease IDs: EFO_0002630, EFO_1001473, HP_0011446, HP_0100543, EFO_0002508 ;
Search terms: prostate, alzheimer, lymphoma
"""
  parser = argparse.ArgumentParser(description='OpenTargets REST API client utility', epilog=epilog)
  ops = ['searchAssociations', 'getEvidence']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input file, gene or disease IDs")
  parser.add_argument("--ids", help="(1) target ID, gene symbol or ENSEMBL; (2) disease ID, EFO_ID; or (3) search term (comma-separated)")
  parser.add_argument("--idtype", choices=IDTYPES, default="disease")
  #parser.add_argument("--source", choices=SOURCES)
  parser.add_argument("--minscore", type=float, help="minimum overall association score")
  parser.add_argument("--nmax", type=int, default=0, help="max results")
  parser.add_argument("--skip", type=int, default=0, help="skip results")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)

  args = parser.parse_args()

  t0 = time.time()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  if args.ofile:
    fout = open(args.ofile,"w+")
  else:
    fout = sys.stdout

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      if line.rstrip(): ids.append(line.rstrip())
    logging.debug('Input queries: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  otclient = opentargets.OpenTargetsClient()

  logging.debug(otclient.get_stats().info)

  if args.op=='searchAssociations':
    if not ids:
      parser.error('--i or --ids required.')
      parser.print_help()
    SearchAssociations(otclient, ids, args.idtype, args.minscore, args.skip, args.nmax, fout)

  elif args.op=='getEvidence':
    if not ids:
      parser.error('--i or --ids required.')
      parser.print_help()
    #Target2Disease_Evidence(otclient, tid, args.did, fout)

  else:
    parser.error('Invalid operation: %s'%args.op)

  logging.info("Elapsed: %s"%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python3
"""
DrugCentral PostgreSql db client.
"""
import os,sys,argparse,re,time,yaml,logging

from .. import drugcentral

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description="DrugCentral PostgreSql client utility", epilog="Search via --ids as regular expressions, e.g.  \"^Alzheimer\"")
  ops = [
	"describe",
	"counts",
	"version",
	"get_structure",
	"get_structure_by_synonym",
	"get_structure_by_indication",
	"get_structure_ids",
	"get_structure_products",
	"get_product",
	"get_product_structures",
	"get_indication_structures",
	"list_products",
	"list_structures",
	"list_structures2smiles",
	"list_structures2molfile",
	"list_active_ingredients",
	"list_indications",
	"search_indications",
	"search_products"
	]
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--i", dest="ifile", help="input ID file")
  parser.add_argument("--ids", help="input IDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--dbhost")
  parser.add_argument("--dbport")
  parser.add_argument("--dbname")
  parser.add_argument("--dbusr")
  parser.add_argument("--dbpw")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.drugcentral.yaml")
  parser.add_argument("--dbschema", default="public")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))


  params={};
  with open(args.param_file, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  if args.dbhost: params['DBHOST'] = args.dbhost 
  if args.dbport: params['DBPORT'] = args.dbport 
  if args.dbname: params['DBNAME'] = args.dbname 
  if args.dbusr: params['DBUSR'] = args.dbusr 
  if args.dbpw: params['DBPW'] = args.dbpw 

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  try:
    dbcon = drugcentral.Utils.Connect(params['DBHOST'], params['DBPORT'], params['DBNAME'], params['DBUSR'], params['DBPW'])
  except Exception as e:
    logging.error("Connect failed.")
    parser.error("{0}".format(str(e)))

  if args.op=='describe':
    drugcentral.Utils.Describe(dbcon, args.dbschema, fout)

  elif args.op=='counts':
    drugcentral.Utils.Counts(dbcon, args.dbschema, fout)

  elif args.op=='version':
    drugcentral.Utils.Version(dbcon, args.dbschema, fout)

  elif args.op=='list_structures':
    drugcentral.Utils.ListStructures(dbcon, args.dbschema, fout)

  elif args.op=='list_structures2smiles':
    drugcentral.Utils.ListStructures2Smiles(dbcon, args.dbschema, fout)

  elif args.op=='list_structures2molfile':
    drugcentral.Utils.ListStructures2Molfile(dbcon, args.dbschema, fout)

  elif args.op=='list_products':
    drugcentral.Utils.ListProducts(dbcon, args.dbschema, fout)

  elif args.op=='list_active_ingredients':
    drugcentral.Utils.ListActiveIngredients(dbcon, args.dbschema, fout)

  elif args.op=='list_indications':
    drugcentral.Utils.ListIndications(dbcon, fout)

  elif args.op=='get_structure':
    drugcentral.Utils.GetStructure(dbcon, ids, fout)

  elif args.op=='get_structure_by_synonym':
    drugcentral.Utils.GetStructureBySynonym(dbcon, ids, fout)

  elif args.op=='get_structure_ids':
    drugcentral.Utils.GetStructureIds(dbcon, ids, fout)

  elif args.op=='get_structure_products':
    drugcentral.Utils.GetStructureProducts(dbcon, ids, fout)

  elif args.op=='get_product_structures':
    drugcentral.Utils.GetProductStructures(dbcon, ids, fout)

  elif args.op=='search_products':
    drugcentral.Utils.SearchProducts(dbcon, ids, fout)

  elif args.op=='search_indications':
    drugcentral.Utils.SearchIndications(dbcon, ids, fout)

  elif args.op=='get_indication_structures':
    drugcentral.Utils.GetIndicationStructures(dbcon, ids, fout)

  else:
    parser.error("Invalid operation: {0}".format(args.op))

  dbcon.close()



#!/usr/bin/env python3
"""
PubMed iCite  REST API client
https://icite.od.nih.gov/api
"""
###
import sys,os,argparse,logging
import urllib.request,json,re
#
from .. import icite

#############################################################################
if __name__=='__main__':
  API_HOST="icite.od.nih.gov"
  API_BASE_PATH="/api/pubs"
  #
  parser = argparse.ArgumentParser(description='PubMed iCite REST API client utility', epilog='Publication metadata.')
  ops = ['get_stats']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", help="PubMed IDs, comma-separated (ex:25533513)")
  parser.add_argument("--i", dest="ifile", help="input file, PubMed IDs")
  parser.add_argument("--nmax", help="list: max to return")
  parser.add_argument("--year", help="list: year of publication")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w", encoding="utf-8") if args.ofile else sys.stdout

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[\s,]+', args.ids.strip())

  if args.op == 'get_stats':
    if not ids: parser.error('Requires PMID[s].')
    icite.Utils.GetStats(BASE_URL, ids, fout)

  else:
    parser.error("Invalid operation: %s"%args.op)




#!/usr/bin/env python3
"""
OpenFDA Adverse Event Reports REST API client.
"""
### 
import sys,os,re,json,argparse,time,yaml,logging

from ... import fda
#
API_HOST='api.fda.gov'
API_BASE_PATH='/drug/event.json'
API_KEY='==SEE $HOME/.fda.yaml=='

NMAX=100

##############################################################################
if __name__=='__main__':
  epilog='''\
Example raw query: "patient.drug.openfda.product_type:HUMAN PRESCRIPTION DRUG"
'''
  parser = argparse.ArgumentParser(description='OpenFDA Adverse Event Reports client', epilog=epilog)
  ops = ['search', 'counts', 'info', 'showfields']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--drug_class", help="EPC pharmacologic class")
  parser.add_argument("--drug_ind", help="drug indication")
  parser.add_argument("--drug_unii", help="drug ID UNII")
  parser.add_argument("--drug_ndc", help="drug ID NDC")
  parser.add_argument("--drug_spl", help="drug ID SPL")
  parser.add_argument("--serious", type=bool, help="serious adverse events")
  parser.add_argument("--fatal", type=bool, help="fatal adverse events (seriousnessdeath)")
  parser.add_argument("--tfrom", help="time-from (received by FDA) (YYYYMMDD)")
  parser.add_argument("--tto", default=time.strftime('%Y%m%d',time.localtime()), help="time-to (received by FDA) (YYYYMMDD)")
  parser.add_argument("--rawquery")
  parser.add_argument("--nmax", type=int, default=NMAX, help="max returned records")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--api_key", default=API_KEY)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  API_BASE_URL='https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout=open(args.ofile,"w+")
    if not fout: parser.error('cannot open: %s'%args.ofile)
  else:
    fout=sys.stdout

  t0=time.time()

  if args.op == "search":
    fda.aer.Utils.Search(args.drug_class, args.drug_ind, args.drug_unii, args.drug_ndc, args.drug_spl, args.tfrom, args.tto, args.serious, args.fatal, args.rawquery, args.NMAX, fout, API_BASE_URL, args.api_key)

  elif args.op == "info":
    rval = fda.aer.Utils.Info(API_BASE_URL)
    for field in rval.keys():
      if field=='results': continue
      print('%16s: %s'%(field,rval[field]))

  elif args.op == "counts":
    print(fda.aer.Utils.GetCounts(API_BASE_URL, args.tfrom, args.tto))

  elif args.op == "showfields":
    fields = fda.aer.Utils.GetFields(API_BASE_URL)
    for field in fields:
      print('\t%s'%field)

  else:
    parser.error("Invalid operation: %s"%args.op)

  logging.info('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))))



#!/usr/bin/env python3
"""	pug_search.py - PubChem/PUG sub|sim|exact search
	Jeremy Yang (2010)
"""
import sys,os,re,urllib,time,argparse,tempfile,logging

from ... import pubchem


def QueryPug_qkey(qkey,webenv,fmt='smiles',gzip=False):
  if gzip: compres='gzip'
  else: compres='none'
  qxml='''\
<PCT-Data><PCT-Data_input><PCT-InputData>
  <PCT-InputData_download><PCT-Download> 
    <PCT-Download_uids><PCT-QueryUids>
      <PCT-QueryUids_entrez><PCT-Entrez>
        <PCT-Entrez_db>pccompound</PCT-Entrez_db>
        <PCT-Entrez_query-key>%(QKEY)s</PCT-Entrez_query-key>
        <PCT-Entrez_webenv>%(WEBENV)s</PCT-Entrez_webenv>
      </PCT-Entrez></PCT-QueryUids_entrez>
    </PCT-QueryUids></PCT-Download_uids>
    <PCT-Download_format value="%(FMT)s"/>
    <PCT-Download_compression value="%(COMPRES)s"/>
  </PCT-Download></PCT-InputData_download>
</PCT-InputData></PCT-Data_input></PCT-Data>
'''%{'QKEY':qkey,'WEBENV':webenv,'FMT':fmt,'COMPRES':compres}
  logging.info('connecting %s...'%PUGURL)
  logging.info('submitting query for qkey: %s ...'%qkey)
  f=pubchem.soap.Utils.UrlOpenTry(PUGURL,qxml)
  pugxml=f.read()
  f.close()
  status,reqid,url,qkey,wenv,error=pubchem.soap.Utils.ParsePugXml(pugxml)
  return status,reqid,url,qkey,wenv,error

#############################################################################
def QueryPug_struct(smi,searchtype='sub',sim_cutoff=0.8,active=False,mlsmr=False,nmax=10000):
  """	 <PCT-CSIdentity value="same-isotope">4</PCT-CSIdentity>
  """
  qxml='''\
<PCT-Data><PCT-Data_input><PCT-InputData><PCT-InputData_query><PCT-Query>
  <PCT-Query_type><PCT-QueryType><PCT-QueryType_css>
    <PCT-QueryCompoundCS>
    <PCT-QueryCompoundCS_query>
      <PCT-QueryCompoundCS_query_data>%(SMILES)s</PCT-QueryCompoundCS_query_data>
    </PCT-QueryCompoundCS_query>
    <PCT-QueryCompoundCS_type>
'''%{'SMILES':smi}
  if searchtype=='sim':
    qxml+='''\
      <PCT-QueryCompoundCS_type_similar><PCT-CSSimilarity>
        <PCT-CSSimilarity_threshold>%(SIM_CUTOFF)d</PCT-CSSimilarity_threshold>
      </PCT-CSSimilarity></PCT-QueryCompoundCS_type_similar>
'''%{'SIM_CUTOFF':sim_cutoff*100}
  elif searchtype=='exact':
    qxml+='''\
      <PCT-QueryCompoundCS_type_identical>
        <PCT-CSIdentity value="same-stereo-isotope">5</PCT-CSIdentity>
      </PCT-QueryCompoundCS_type_identical>
'''
  else:
    qxml+='''\
      <PCT-QueryCompoundCS_type_subss><PCT-CSStructure>
        <PCT-CSStructure_isotopes value="false"/>
        <PCT-CSStructure_charges value="false"/>
        <PCT-CSStructure_tautomers value="false"/>
        <PCT-CSStructure_rings value="false"/>
        <PCT-CSStructure_bonds value="true"/>
        <PCT-CSStructure_chains value="true"/>
        <PCT-CSStructure_hydrogen value="false"/>
      </PCT-CSStructure></PCT-QueryCompoundCS_type_subss>
'''
  qxml+='''\
    </PCT-QueryCompoundCS_type>
    <PCT-QueryCompoundCS_results>%(NMAX)d</PCT-QueryCompoundCS_results>
    </PCT-QueryCompoundCS>
  </PCT-QueryType_css></PCT-QueryType>
'''%{'NMAX':nmax}
  if active:
    qxml+='''\
  <PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
    <PCT-QueryCompoundEL_activity>
      <PCT-CLByBioActivity value="active">2</PCT-CLByBioActivity>
    </PCT-QueryCompoundEL_activity>
  </PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
'''
  if mlsmr:
    qxml+='''\
  <PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
    <PCT-QueryCompoundEL_source><PCT-CLByString>
      <PCT-CLByString_qualifier value="must">1</PCT-CLByString_qualifier>
      <PCT-CLByString_category>MLSMR</PCT-CLByString_category>
    </PCT-CLByString></PCT-QueryCompoundEL_source>
  </PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
'''
  qxml+='''\
  </PCT-Query_type></PCT-Query>
</PCT-InputData_query></PCT-InputData></PCT-Data_input></PCT-Data>
'''
  logging.info('connecting %s...'%PUGURL)
  if sim:
    logging.info('query (%f): %s ...'%(sim_cutoff,smiles))
  else:
    logging.info('substructure query: %s ...'%smiles)
  f=pubchem.soap.Utils.UrlOpenTry(PUGURL,qxml)
  pugxml=f.read()
  f.close()

  status,reqid,url,qkey,wenv,error=pubchem.soap.Utils.ParsePugXml(pugxml)
  return status,reqid,qkey,wenv,error

#############################################################################
def ElapsedTime(t_start):
  t_elapsed=time.time()-t_start
  tsec=t_elapsed%60
  tmin=t_elapsed/60
  return ("%d:%02d"%(tmin,tsec))

#############################################################################
if __name__=='__main__':
  PROG=os.path.basename(sys.argv[0])
  PUGURL='http://pubchem.ncbi.nlm.nih.gov/pug/pug.cgi'
  POLL_WAIT=10
  MAX_WAIT=300
  SIM_CUTOFF=0.80

  parser = argparse.ArgumentParser(description="PubChem PUG SOAP client: sub|sim|exact search, fetch smiles|sdfs")
  parser.add_argument("--o", dest="ofile", help="output smiles|sdf file (w/ CIDs)")
  parser.add_argument("--qsmi", help="input query smiles (or smarts)")
  parser.add_argument("--fmt", help="smiles|sdf [default via filename]")
  parser.add_argument("--gz", help="output gzipped [default via filename]")
  parser.add_argument("--out_cids", dest="out_cids_file", help="output CIDs (only w/ smiles output)")
  parser.add_argument("--similarity", help="similarity search [default is substructure]")
  parser.add_argument("--exact", help="exact search [default is substructure]")
  parser.add_argument("--sim_cutoff", type=float, default=SIM_CUTOFF, help="similarity cutoff, Tanimoto")
  parser.add_argument("--active", help="active mols only (in any assay)")
  parser.add_argument("--mlsmr", help="MLSMR mols only")
  parser.add_argument("--nmax", type=int, help="max hits returned")
  parser.add_argument("--max_wait", type=int, default=MAX_WEIGHT, help="max wait for query")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  if not args.qsmi:
    parser.error('smiles required\n'+usage)
  logging.info('query: "%s"'%args.qsmi)

  if args.ofile:
    fout=open(args.ofile,'wb')
    if (args.ofile[-3:]=='.gz'):
      gzip=True
      ext=re.sub('^.*\.','',args.ofile[:-3])
    else:
      ext=re.sub('^.*\.','',args.ofile)
    if not args.fmt: fmt=ext
    if fmt=='smi': fmt='smiles'
    if fmt not in ('smiles','sdf'):
      parser.error('format "smiles" or "sdf" required')

  if args.out_cids_file:
    if fmt=='sdf':
      parser.error('--out_cids not allowed with SDF output.')

    fout_cids=open(args.out_cids_file,'wb')

  t0=time.time()	## start timer

  status,reqid,qkey,wenv,error=QueryPug_struct(args.qsmi,args.searchtype,args.sim_cutoff,args.active,args.mlsmr,args.nmax)

  if status not in ('success','queued','running'):
    parser.error('query failed; quitting (status=%s,error="%s").'%(status,error))
  logging.info('query status: %s'%status)

  if not reqid and not qkey:
    parser.error('query ok but no ID; quitting.')

  i_poll=0
  while not qkey:
    if time.time()-t0>args.max_wait:
      pubchem.soap.Utils.PollPug(reqid,'cancel')
      parser.error('max wait exceeded (%d sec); quitting.'%args.max_wait)
    logging.info('polling PUG [ID=%d]...'%reqid)
    status,reqid_new,url,qkey,wenv,error=pubchem.soap.Utils.PollPug(reqid,'status')
    if reqid_new: reqid=reqid_new
    if qkey: break
    logging.info('%s elapsed; %d sec wait; status=%s'%(ElapsedTime(t0),POLL_WAIT,status))
    logging.info('%s'%(error))
    if re.search('No result found',error,re.I):
      parser.error(error)
    time.sleep(POLL_WAIT)
    i_poll+=1


  status,reqid,url,qkey,wenv,error=QueryPug_qkey(qkey,wenv,fmt,gzip)
  i_poll=0
  while not url:
    if time.time()-t0>args.max_wait:
      pubchem.soap.Utils.PollPug(reqid,'cancel')
      parser.error('max wait exceeded (%d sec); quitting.'%args.max_wait)
    logging.info('polling PUG [ID=%d]...'%reqid)
    status,reqid,url,qkey,wenv,error=pubchem.soap.Utils.PollPug(reqid,'status')
    if url: break
 
    logging.info('%s elapsed; %d sec wait; status=%s'%(ElapsedTime(t0),POLL_WAIT,status))
    time.sleep(POLL_WAIT)
    i_poll+=1

  logging.info('query elapsed time: %s'%(ElapsedTime(t0)))
  if args.ofile:
    logging.info('URL: %s'%url)
    logging.info('downloading to %s...'%args.ofile)

  fd,tmpfile=tempfile.mkstemp('.txt', PROG)
  f=os.fdopen(fd,"w")
  nbytes=pubchem.soap.Utils.DownloadUrl(url,f)
  f.close()

  if fmt=='smiles':
    ftmp=open(tmpfile)
    n=0
    while True:
      line=ftmp.readline()
      if not line: break
      fields=line.split()
      if args.ofile: fout.write("%s %s\n"%(fields[1],fields[0]))
      if args.out_cids_file: fout_cids.write("%s\n"%(fields[0]))
      n+=1
    logging.info('compounds: %d'%(n))
    ftmp.close()
  else:
    os.system("cp %s %s"%(tmpfile,args.ofile))
  os.unlink(tmpfile)

  sys.stderr.write('%s: format: %s'%(PROG,fmt))
  if args.gzip: sys.stderr.write('(gzipped)')
  sys.stderr.write('\n')

  if args.ofile:
    fout.close()
    nbytes=os.stat(args.ofile).st_size
    logging.info('%s (%.2fMB)'%(args.ofile,nbytes/1e6))

  if args.out_cids_file:
    fout_cids.close()
    nbytes=os.stat(args.out_cids_file).st_size
    logging.info('%s (%.2fMB)'%(args.out_cids_file,nbytes/1e6))




#!/usr/bin/env python3
"""
Utility functions for the PubChem PUG REST API.

* https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest

	<DOMAIN>/<NAMESPACE>/<IDENTIFIERS>

	compound/{cid|name|smiles|inchi|inchikey|listkey}
	compound/substructure/{smiles|inchi}
	substance/{sid|name|listkey}
	substance/sourceid/<source name> 
	substance/sourceall/<source name> 
	substance/sid/<SID>/cids/JSON
	substance/sid/<SID>/cids/XML?cids_type=all
	assay/{aid|listkey} 
 	assay/type/{all|confirmatory|doseresponse|onhold|panel|rnai|screening|summary}
 	assay/sourceall/<source name>
	sources/{substance|assay}

  <IDENTIFIERS> = comma-separated list of positive integers (cid, sid, aid) or
  strings (source, inchikey, listkey); single identifier string
  (name, smiles; inchi by POST only)
"""
###
import sys,os,io,re,csv,json,pandas,math,time,logging
import urllib.request,urllib.parse
#
from ..util import rest_utils
#
OUTCOME_CODES = {
        'inactive':1,
        'active':2,
        'inconclusive':3,
        'unspecified':4,
        'probe':5}
#
#############################################################################
def ListSources(base_url, src_type, fout):
  rval = rest_utils.GetURL(base_url+'/sources/%s/JSON'%src_type, parse_json=True)
  logging.debug(json.dumps(rval,indent=2))
  sources = rval['InformationList']['SourceName'] if 'InformationList' in rval and 'SourceName' in rval['InformationList'] else []
  n_src=0;
  for source in sorted(sources):
    fout.write('"%s"\n'%source)
    n_src+=1
  logging.info('n_src (%s): %d'%(src_type, n_src))

##############################################################################
def GetSID2SDF(base_url, id_query):
  url=(base_url+'/substance/sid/%d/SDF'%id_query)
  txt=rest_utils.GetURL(url)
  return txt

#############################################################################
def GetSID2CID(base_url, sids, fout):
  cids = set()
  if fout: fout.write("SID\tCID\n")
  for sid in sids:
    rval = rest_utils.GetURL(base_url+'/substance/sid/%s/cids/JSON?cids_type=standardized'%sid, parse_json=True)
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      cids_this = info['CID'] if 'CID' in info else []
      for cid in cids_this:
        cids.add(str(cid))
        if fout: fout.write("%s\t%s\n"%(sid, cid))
  return list(cids)

#############################################################################
def GetSID2Smiles(base_url, sids, fout):
  cids = GetSID2CID(base_url, sids, None)
  GetCID2Smiles(base_url, cids, fout)

#############################################################################
def GetCID2SID(base_url, cids, fout):
  sids = set()
  if fout: fout.write("CID\tSID\n")
  for cid in cids:
    rval = rest_utils.GetURL(base_url+'/compound/cid/%s/sids/JSON'%cid, parse_json=True)
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      sids_this = info['SID'] if 'SID' in info else []
      for sid in sids_this:
        sids.add(str(sid))
        if fout: fout.write("%s\t%s\n"%(cid, sid))
  return list(sids)

#############################################################################
### Must quote forward slash '/' in smiles.  2nd param in quote()
### specifies "safe" chars to not quote, '/' being the default, so specify
### '' for no safe chars.
#############################################################################
def GetSmiles2CID(base_url, smis, fout):
  n_out=0;
  fout.write("CID\tSMILES\tName\n")
  for smi in smis:
    name = re.sub(r'^[\S]+\s', '', smi) if re.search(r'^[\S]+\s', smi) else ""
    smi = re.sub(r'\s.*$', '', smi)
    rval = rest_utils.GetURL(base_url+'/compound/smiles/%s/cids/JSON'%urllib.parse.quote(smi, ''), parse_json=True)
    if not rval: continue
    cids = rval['IdentifierList']['CID'] if 'IdentifierList' in rval and 'CID' in rval['IdentifierList'] else []
    for cid in cids:
      fout.write("%s\t%s\t%s\n"%(cid, smi, name))
      n_out+=1
  logging.info('SMIs: %d; CIDs out: %d'%(len(smis), n_out))

#############################################################################
def GetCID2AssaySummary(base_url, ids, fout):
  """Example CIDs: 2519 (caffeine), 3034034 (quinine)"""
  n_out=0;
  for i,id_this in enumerate(ids):
    rval = rest_utils.GetURL(base_url+'/compound/cid/%s/assaysummary/CSV'%id_this)
    logging.debug(rval)
    if not rval: continue
    df = pandas.read_csv(io.StringIO(rval), sep=',')
    df.to_csv(fout, sep='\t', index=False, header=bool(i==0))
    n_out += df.shape[0]
  logging.info('CIDs: %d; assay summaries out: %d'%(len(ids), n_out))

#############################################################################
def GetSID2AssaySummary(base_url, ids, fout):
  n_out=0;
  for id_this in ids:
    rval = rest_utils.GetURL(base_url+'/substance/sid/%s/assaysummary/CSV'%id_this)
    if not rval: continue
    df = pandas.read_csv(io.StringIO(rval), sep=',')
    df.to_csv(fout, sep='\t', index=False, header=bool(i==0))
    n_out += df.shape[0]
  logging.info('SIDs: %d; assay summaries out: %d'%(len(ids), n_out))

#############################################################################
def GetCID2Properties(base_url, ids, fout):
  PROPTAGS = [ "CanonicalSMILES", "IsomericSMILES", "InChIKey", "InChI", "MolecularFormula", "HeavyAtomCount", "MolecularWeight", "XLogP", "TPSA"]
  ids_dict = {'cid':(','.join(map(lambda x:str(x), ids)))}
  url = (base_url+'/compound/cid/property/%s/CSV'%(','.join(PROPTAGS)))
  rval = rest_utils.PostURL(url, headers={'Accept':'text/CSV', 'Content-type':'application/x-www-form-urlencoded'}, data=ids_dict)
  df = pandas.read_csv(io.StringIO(rval), sep=',')
  df.to_csv(fout, sep='\t', index=False)
  logging.info("Input IDs: {0}; Output records: {1}".format(len(ids), df.shape[0]))

#############################################################################
def GetCID2Inchi(base_url, ids, fout):
  PROPTAGS = [ "InChIKey", "InChI" ]
  ids_dict = {'cid':(','.join(map(lambda x:str(x), ids)))}
  url = (base_url+'/compound/cid/property/%s/CSV'%(','.join(PROPTAGS)))
  rval = rest_utils.PostURL(url, headers={'Accept':'text/CSV', 'Content-type':'application/x-www-form-urlencoded'}, data=ids_dict)
  df = pandas.read_csv(io.StringIO(rval), sep=',')
  df.to_csv(fout, sep='\t', index=False)
  logging.info("Input IDs: {0}; Output InChIs: {1}".format(len(ids), df.shape[0]))

##############################################################################
#def GetCID2SDF(base_url, ids, fout):
#  n_out=0;
#  for cid in ids:
#    url = (base_url+'/compound/cid/%d/SDF'%cid)
#    rval = rest_utils.GetURL(url)
#    fout.write(rval)
#  logging.info('SDFs out: %d'%n_out)
#
#############################################################################
def GetCID2SDF(base_url, ids, fout):
  """Faster via POST(?). Request in chunks.  Works for 50, and not
  for 200 (seems to be a limit)."""
  nchunk=50; nskip_this=0; n_out=0;
  while True:
    if nskip_this>=len(ids): break
    idstr = (','.join(map(lambda x:str(x), ids[nskip_this:nskip_this+nchunk])))
    rval = rest_utils.PostURL(base_url+'/compound/cid/SDF', data={'cid':idstr})
    fout.write(rval)
    n_out += len(re.findall(r'^\$\$\$\$$', rval, re.M))
    nskip_this+=nchunk
  logging.info('SDFs out: %d'%n_out)

#############################################################################
def GetSID2SDF(base_url, sids, skip, nmax, fout):
  """Faster via POST(?). Request in chunks.  Works for 50, and not
  for 200 (seems to be a limit)."""
  n_out=0;
  if skip: logging.debug('skip: [1-%d]'%skip)
  nchunk=50; nskip_this=skip;
  while True:
    if nskip_this>=len(sids): break
    nchunk = min(nchunk, nmax-(nskip_this-skip))
    n_sid_in+=nchunk
    idstr = (','.join(map(lambda x:str(x), sids[nskip_this:nskip_this+nchunk])))
    rval = rest_utils.PostURL(base_url+'/substance/sid/SDF', data={'sid':idstr})
    if rval:
      fout.write(rval)
      n_out += len(re.findall(r'^\$\$\$\$$', rval, re.M))
    nskip_this+=nchunk
    if nmax and (nskip_this-skip)>=nmax:
      logging.info('NMAX limit reached: %d'%nmax)
      break
  logging.info('SIDs: %d ; SDFs out: %d'%(len(sids), n_out))

#############################################################################
def GetCID2Smiles(base_url, ids, isomeric, fout):
  """Returns Canonical or Isomeric SMILES."""
  t0 = time.time()
  nchunk=50; nskip_this=0;
  n_in=0; n_out=0; n_err=0;
  fout.write('CID\t%s\n'%('IsomericSMILES' if isomeric else 'CanonicalSMILES'))
  while True:
    if nskip_this>=len(ids): break
    ids_this = ids[nskip_this:nskip_this+nchunk]
    n_in+=len(ids_this)
    idstr = (','.join(map(lambda x:str(x), ids_this)))
    prop = 'IsomericSMILES' if isomeric else 'CanonicalSMILES'
    txt = rest_utils.PostURL(base_url+'/compound/cid/property/%s/CSV'%prop, data={'cid':idstr})
    if not txt:
      n_err+=1
      break
    lines = txt.splitlines()
    for line in lines:
      cid,smi = re.split(',', line)
      if cid.upper()=='CID': continue #header
      try:
        cid = int(cid)
      except:
        continue
      smi = smi.replace('"', '')
      fout.write('%s %d\n'%(smi, cid))
      n_out+=1
    nskip_this+=nchunk
    if (n_in%1000)==0:
      logging.info('processed: %6d / %6d (%.1f%%); elapsed time: %s'%(n_in, len(ids), 100.0*n_in/len(ids), time.gmtime(time.time()-t0)))
  logging.info('CIDs in: %d; SMILES (isomeric=%s) out: %d; errors: %d'%(n_in, bool(isomeric), n_out, n_err))

#############################################################################
def Inchi2CID(base_url, inchis, fout):
  '''	Must be POST with "Content-Type: application/x-www-form-urlencoded"
	or "Content-Type: multipart/form-data" with the POST body formatted accordingly.
	See: http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html and
	http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST_Tutorial.html
  '''
  n_out=0;
  fout.write("InChI\tCID\n")
  for inchi in inchis:
    url = "http://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/inchi/cids/TXT"
    logging.info('inchi="%s"'%(inchi))
    body_dict={'inchi':urllib.parse.quote(inchi)}
    rval = rest_utils.PostURL(url=url, headers={'Content-Type':'application/x-www-form-urlencoded','Accept':'text/plain'}, data=body_dict)
    cids_this = set()
    lines = rval.splitlines()
    for line in lines:
      if re.match(r'[\d]+$', line.strip()):
        cid = line.strip()
        cids_this.add(cid)
        fout.write("%s\t%s\n"%(inchi, cid))
        n_out+=1
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetAssayName(base_url, aids, fout):
  for aid in aids:
    xmlstr = rest_utils.GetURL(base_url+'/assay/aid/%d/description/XML'%id_query)
    name, source = AssayXML2NameAndSource(xmlstr)
    fout.write('%d\t%s\t%s'%(aid, name, source))

#############################################################################
def GetAssayDescriptions(base_url, ids, skip, nmax, fout):
  """Example AIDs: 159014"""
  n_in=0; n_out=0;
  fout.write("aid\tname\tsource\trevision\tassay_group\tproject_category\tactivity_outcome_method\tdescription\tcomment\n")
  for aid in ids:
    n_in+=1
    if skip and n_in<skip: continue
    if nmax and n_out==nmax: break
    url = (base_url+'/assay/aid/%s/description/JSON'%aid)
    rval = rest_utils.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2))
    assays = rval['PC_AssayContainer'] if 'PC_AssayContainer' in rval else []
    for assay in assays:
      descr = assay['assay']['descr'] if 'assay' in assay and 'descr'  in assay['assay'] else None
      if not descr: continue
      #aid = str(descr['aid']['id'])
      name = descr['name']
      source = descr['aid_source']['db']['name']
      revision = str(descr['revision'])
      project_category = str(descr['project_category'])
      activity_outcome_method = str(descr['activity_outcome_method'])
      assay_group = descr['assay_group'] #list of PMIDs
      description_lines = descr['description'] #list of lines
      comment_lines = descr['comment'] #list of lines
      description = (r'\\n').join(description_lines)
      comment = (r'\\n').join(comment_lines)
      vals = [aid, name, source, revision, (';'.join(assay_group)), project_category, activity_outcome_method, description, comment]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
  logging.info("n_out: %d"%n_out)

#############################################################################
def OutcomeCode(txt):
  return OUTCOME_CODES[txt.lower()] if txt.lower() in OUTCOME_CODES else OUTCOME_CODES['unspecified']

#############################################################################
def GetAssaySIDResults(base_url, aids, sids, skip, nmax, fout):
  '''One CSV line for each activity.  skip and nmax applied to AIDs.
In this version of the function, use the "concise" mode to download full data
for each AID, then iterate through SIDs and use local hash.
'''
  if skip: logging.info('skip: [1-%d]'%skip)
  if nmax: logging.info('NMAX: %d'%nmax)
  n_aid_in=0; n_aid_done=0; n_out=0;
  nchunk_sid=100;
  #fout.write('aid,sid,outcome,version,rank,year\n')
  tags = None
  for aid in aids:
    n_aid_in+=1
    n_out_this=0;
    if skip and n_aid_in<skip: continue
    if nmax and n_aid_done>=nmax: break
    n_aid_done+=1
    logging.debug('Request: (%d) [AID=%d] SID count: %d'%(n_aid_done,aid,len(sids)))
    try:
      url = base_url+'/assay/aid/%d/concise/JSON'%(aid)
      logging.debug('%s'%(url))
      rval = rest_utils.GetURL(url,parse_json=True)
    except Exception as e:
      logging.info('Error: %s'%(e))
      break

    if not type(rval) is dict:
      #logging.info('Error: %s'%(str(rval)))
      f = open("z.z", "w")
      f.write(str(rval))
      f.close()
      logging.debug('rval not DictType.  See z.z')
      break
    adata = {} # data this assay
    tags_this = rval["Table"]["Columns"]["Column"]
    logging.debug('tags_this = %s'%(str(tags_this)))
    j_sid = None;
    j_res = None;
    if not tags:
      tags = tags_this
      for j,tag in enumerate(tags):
        if tag == 'SID': j_sid = j
        if tag == 'Bioactivity Outcome': j_res = j
      fout.write(','.join(tags)+'\n')

    rows = rval["Table"]["Row"]

    for row in rows:
      cells = row["Cell"]
      if len(cells) != len(tags):
        logging.info('Error: n_cells != n_tags (%d != %d)'%(len(cells), len(tags)))
        continue

      sid = cells[j_sid]
      res = cells[j_res]
      adata[sid] = res

      #break #DEBUG

#############################################################################
def GetAssaySIDResults_OLD(base_url, aids, sids, skip, nmax, fout):
  '''One CSV line for each activity.  skip and nmax applied to AIDs.
Must chunk the SIDs, since requests limited to 10k SIDs.
'''
  if skip: logging.info('skip: [1-%d]'%skip)
  if nmax: logging.info('NMAX: %d'%nmax)
  n_aid_in=0; n_aid_done=0; n_out=0;
  nchunk_sid=100;
  fout.write('aid,sid,outcome,version,rank,year\n')
  for aid in aids:
    n_aid_in+=1
    n_out_this=0;
    if skip and n_aid_in<skip: continue
    if nmax and n_aid_done>=nmax: break
    n_aid_done+=1
    logging.debug('Request: (%d) [AID=%d] SID count: %d'%(n_aid_done,aid,len(sids)))
    nskip_sid=0;
    while True:
      if nskip_sid>=len(sids): break
      logging.debug('(%d) [AID=%s] ; SIDs [%d-%d of %d] ; measures: %d'%(n_aid_done,aid,nskip_sid+1,min(nskip_sid+nchunk_sid,len(sids)),len(sids),n_out_this))
      sidstr=(','.join(map(lambda x:str(x),sids[nskip_sid:nskip_sid+nchunk_sid])))
      try:
        rval = rest_utils.GetURL(base_url+'/assay/aid/%d/JSON?sid=%s'%(aid,sidstr),parse_json=True)
      except Exception as e:
        logging.info('Error: %s'%(e))
        break
      if not type(rval) is dict:
        logging.info('Error: %s'%(str(rval)))
        break
      pcac = rval['PC_AssayContainer'] if 'PC_AssayContainer' in rval else None
      if len(pcac)<1:
        logging.info('Error: [%s] empty PC_AssayContainer.'%aid)
        break
      if len(pcac)>1:
        logging.info('NOTE: [%s] multiple assays in PC_AssayContainer.'%aid)
      for assay in pcac:
        if not assay: continue
        logging.debug(json.dumps(assay,indent=2))
        ameta = assay['assay'] if 'assay' in assay else None
        adata = assay['data'] if 'data' in assay else None
        if not ameta: logging.info('Error: [%s] no metadata.'%aid)
        if not adata: logging.info('Error: [%s] no data.'%aid)
        for datum in adata:
          sid = datum['sid'] if 'sid' in datum else ''
          outcome = datum['outcome'] if 'outcome' in datum else ''
          version = datum['version'] if 'version' in datum else ''
          rank = datum['rank'] if 'rank' in datum else ''
          date = datum['date'] if 'date' in datum else {}
          year = date['std']['year'] if 'std' in date and 'year' in date['std'] else ''
          fout.write('%d,%s,%d,%s,%s,%s\n'%(aid,sid,OutcomeCode(outcome),version,rank,year))
          fout.flush()
          n_out_this+=1
      nskip_sid+=nchunk_sid
    logging.debug('Result: [AID=%d] total measures: %d'%(aid,n_out_this))

  return n_aid_in,n_out

#############################################################################
def AssayXML2NameAndSource(xmlstr):
  '''Required: xpath - XPath Queries For DOM Trees, http://py-dom-xpath.googlecode.com/'''
  tag_name='PC-AssayDescription_name'
  tag_source='PC-DBTracking_name'

  #import xpath #old
  #dom=xml.dom.minidom.parseString(xmlstr)
  #name=xpath.findvalue('//%s'%tag_name,dom)  ##1st
  #source=xpath.findvalue('//%s'%tag_source,dom)  ##1st

  from xml.etree import ElementTree #newer

  root = ElementTree.fromstring(xmlstr)
  name=root.find('//%s'%tag_name).text  ##1st
  source=root.find('//%s'%tag_sourcedom).text  ##1st

  return name,source

#############################################################################
def GetSID2Synonyms(base_url, sids, fout):
  synonyms=[]
  if fout: fout.write("SID\tSynonym\n")
  for sid in sids:
    rval = rest_utils.GetURL(base_url+'/substance/sid/%s/synonyms/JSON'%(sid), parse_json=True)
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      synonyms_this = info['Synonym'] if 'Synonym' in info else []
      for synonym in synonyms_this:
        synonyms.append(synonym)
        if fout: fout.write("%s\t%s\n"%(sid,synonym))
  return synonyms

#############################################################################
def GetCID2Synonyms(base_url, cids, skip, nmax, nmax_per_cid, fout):
  fout.write('CID\tSynonym\n')
  sids_all = set([])
  i_cid=0; n_out=0;
  for cid in cids:
    i_cid+=1
    if skip and i_cid<=skip: continue
    sids_this = GetCID2SID(base_url, [cid], None)
    sids_all |= set(sids_this)
    synonyms_this_cid = set()
    for sid in sids_this:
      synonyms_this_sid = GetSID2Synonyms(base_url, [sid], None)
      synonyms_this_cid |= set(synonyms_this_sid)
    synonyms_this_cid_nice = SortCompoundNamesByNiceness(list(synonyms_this_cid))
    for j,synonym in enumerate(synonyms_this_cid_nice):
      if nmax_per_cid and j>=nmax_per_cid:
        logging.info('%d. CID=%s: synonyms out+truncated=all: %d+%d=%d'%(
            i_cid, cid, nmax_per_cid,
            len(synonyms_this_cid_nice)-nmax_per_cid,
            len(synonyms_this_cid_nice)))
        break
      fout.write('%s\t%s\n'%(cid, synonym))
      n_out+=1
    logging.info('%d. CID=%s: SIDs: %d ; synonyms: %d (%d)'%(i_cid, cid,
        len(sids_this), len(synonyms_this_cid),
        min(len(synonyms_this_cid_nice), nmax_per_cid)))
    if nmax and i_cid>=(skip+nmax): break
  logging.info('Totals: CIDs: %d; SIDs: %d; Synonyms: %d'%(len(cids), len(sids_all), n_out))

#############################################################################
def GetName2SID(base_url, names, fout):
  n_sid=0; sids_all = set()
  if fout: fout.write("Name\tSID\n")
  for name in names:
    rval = rest_utils.GetURL(base_url+'/substance/name/%s/sids/JSON'%urllib.parse.quote(name), parse_json=True)
    sids_this = rval['IdentifierList']['SID']
    for sid in sids_this:
      n_sid+=1
      if fout: fout.write("%s\t%s\n"%(name, sid))
      sids_all |= set(sids_this)
  logging.info("n_name: %d; n_sid: %d; n_unique: %d"%(len(names), n_sid, len(sids_all)))
  return list(sids_all)

#############################################################################
def GetName2CID(base_url, names, fout):
  n_cid=0; cids_all = set()
  if fout: fout.write("Name\tCID\n")
  for name in names:
    sids_this = GetName2SID(base_url, [name], None)
    cids_this = list(set(GetSID2CID(base_url, sids_this, None)) )
    for cid in cids_this:
      n_cid+=1
      if fout: fout.write("%s\t%s\n"%(name, cid))
    cids_all |= set(cids_this)
  logging.info("n_name: %d; n_cid: %d; n_unique: %d"%(len(names), n_cid, len(cids_all)))
  return list(cids_all)

#############################################################################
def GetName2Synonyms(base_url, names, fout):
  n_synonym=0; sids_all = set(); synonyms_all = set()
  if fout: fout.write("Name\tSID\tSynonym\n")
  for name in names:
    sids_this = GetName2SID(base_url, [name], None)
    sids_all |= set(sids_this)
    for sid in sids_this:
      rval = rest_utils.GetURL(base_url+'/substance/sid/%s/synonyms/JSON'%(sid), parse_json=True)
      infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
      synonyms_this_sid = set()
      for info in infos:
        synonyms_this = info['Synonym'] if 'Synonym' in info else []
        synonyms_this_sid |= set(synonyms_this)
        for synonym in synonyms_this:
          n_synonym+=1
          if fout: fout.write("%s\t%s\t%s\n"%(name, sid, synonym))
      synonyms_all |= set(synonyms_this_sid)
  logging.info("n_name: %d; n_synonym: %d; n_unique: %d; n_sids_unique: %d"%(len(names), n_synonym, len(synonyms_all), len(sids_all)))

#############################################################################
def NameNicenessScore(name, len_optimal=9):
  """Heuristic for human comprehensibility. In PubChem, a 'name' may be
any ID, often not useful."""
  score=0;
  pat_proper = re.compile(r'^[A-Z][a-z]+$')
  pat_text = re.compile(r'^[A-Za-z ]+$')
  if pat_proper.match(name): score+=100
  elif pat_text.match(name): score+=50
  elif re.match(r'^[A-z][A-z][A-z][A-z][A-z][A-z][A-z].*$', name): score+=10
  if re.search(r'[\[\]\{\}]', name): score-=50
  if re.search(r'[_/\(\)]', name): score-=10
  if re.search(r'\d', name): score-=10
  score -= math.fabs(len_optimal-len(name))
  return score

#############################################################################
def SortCompoundNamesByNiceness(names):
  names_scored = {n:NameNicenessScore(n) for n in names}
  names_ranked = [[score,name] for name,score in names_scored.items()]
  names_ranked.sort()
  names_ranked.reverse()
  return [name for score,name in names_ranked]

#############################################################################
def GetCpdAssayStats(base_url, cid, smiles, aidset, fout_mol, fout_act, aidhash):
  """
THIS FUNCTION NEEDS WORK.
From PubChem PUG REST API, determine activity stats for compounds.
 
 Input: smiles and CIDs
 Output (1):
   SMILES CID SID aTested aActive wTested wActive 
     where:
   aTested - how many assays where that cpd has been tested
   aActive - how many assays where that cpd has been tested active
   wTested - how many samples (wells) with that cpd have been tested
   wActive - how many samples (wells) with that cpd have been tested active
 
 Output (2): activity file, each line:
   CID SID AID Outcome
     where:
   1 = inactive
   2 = active
   3 = inconclusive
   4 = unspecified
   5 = probe
   multiple, differing 1, 2 or 3 = discrepant
   not 4 = tested
"""
  aids_tested=set(); aids_active=set();
  sids_tested=set(); sids_active=set();
  n_sam=0; n_sam_active=0; mol_active=False; mol_found=False;

  try:
    fcsv=rest_utils.GetURL(base_url+'/compound/cid/%d/assaysummary/CSV'%cid)
  except Exception as e:
    logging.error('[%d] REST request failed; %s'%(cid, e))
    fout_mol.write("%s %d\n"%(smiles, cid))
    return mol_found,mol_active,n_sam,n_sam_active
  mol_found=True

  try:
    csvReader=csv.DictReader(fcsv.splitlines(),fieldnames=None,restkey=None,restval=None,dialect='excel',delimiter=',',quotechar='"')
    csvrow=csvReader.next()    ## must do this to get fieldnames
  except Exception as e:
    logging.error('[CID=%d] CSV problem:%s'%(cid,e))
    fout_mol.write("%s %d\n"%(smiles,cid))
    return mol_found,mol_active,n_sam,n_sam_active

  for field in ('AID','CID','SID','Bioactivity Outcome','Bioassay Type'):
    if field not in csvReader.fieldnames:
      logging.error('[CID=%d] bad CSV header, no "%s" field.'%(cid,field))
      logging.debug('DEBUG: fieldnames: %s'%(','.join(map((lambda
x:'"%s"'%x),csvReader.fieldnames))))
      fout_mol.write("%s %d\n"%(smiles,cid))
      return mol_found,mol_active,n_sam,n_sam_active

  n_in=0
  while True:
    try:
      csvrow=csvReader.next()
    except:
      break  ## EOF
    n_in+=1

    try:
      aid=int(csvrow['AID'])
      sid=int(csvrow['SID'])
      csvrow_cid=int(csvrow['CID'])
    except:
      logging.error('[CID=%d] bad CSV line; problem parsing: "%s"'%(cid,str(csvrow)))
      continue
    if cid!=csvrow_cid:
      logging.error('Aaack! [CID=%d] CID mismatch: != %d'%(cid,csvrow_cid))
      return mol_found,mol_active,n_sam,n_sam_active

    if aidset:
      if aid not in aidset:
        continue
      #logging.debug('DEBUG: AID [%d] ok (pre-selected).'%(aid))

    ## Assay filtering done; now update statistics (sTested, sActive).
    n_sam+=1
    aidhash[aid]=True
    act_outcome=csvrow['Bioactivity Outcome'].lower()
    if act_outcome in OUTCOME_CODES:
      fout_act.write("%d,%d,%d,%d\n"%(cid,sid,aid,OUTCOME_CODES[act_outcome]))
    else:
      logging.debug('[%d] unrecognized outcome (CID=%d,AID=%d): "%s"'%(n_in,cid,aid,act_outcome))

    aids_tested.add(aid)
    sids_tested.add(sid)
    if act_outcome=='active':
      n_sam_active+=1
      mol_active=True
      aids_active.add(aid)
      sids_active.add(sid)

  logging.debug('[CID=%d] CSV data lines: %d'%(cid,n_in))

  fout_mol.write("%s %d %d %d %d %d %d %d\n"%(
	smiles,
	cid,
	len(sids_tested),
	len(sids_active),
	len(aids_tested),
	len(aids_active),
	n_sam,
	n_sam_active))
  logging.debug((
	"cid=%d,sTested=%d,sActive=%d,aTested=%d,aActive=%d,wTested=%d,wActive=%d"%(
	cid,
	len(sids_tested),
	len(sids_active),
	len(aids_tested),
	len(aids_active),
	n_sam,
	n_sam_active)))

  return mol_found,mol_active,n_sam,n_sam_active

#############################################################################
def GetCpdAssayData(base_url, cid_query, aidset, fout):
  """This function needs fixing."""
  try:
    fcsv=rest_utils.GetURL(base_url+'/compound/cid/%d/assaysummary/CSV'%cid_query)
  except Exception as e:
    logging.error('[%d] REST request failed; %s'%(cid_query,e))
    return False

  if not fcsv:
    return False

  try:
    csvReader=csv.DictReader(fcsv.splitlines(),fieldnames=None,restkey=None,restval=None,dialect='excel',delimiter=',',quotechar='"')
    csvrow=csvReader.next()    ## must do this to get fieldnames
  except Exception as e:
    logging.error('[CID=%d] CSV problem: %s'%(cid_query,e))
    return True

  for field in ('CID','SID','AID','Bioactivity Outcome','Activity Value [uM]'):
    if field not in csvReader.fieldnames:
      logging.error('[CID=%d] bad CSV header, no "%s" field.'%(cid_query,field))
      logging.debug('fieldnames: %s'%(','.join(map((lambda x:'"%s"'%x),csvReader.fieldnames))))
      return True

  n_in=0; n_act=0;
  while True:
    try:
      csvrow=csvReader.next()
    except:
      break  ## EOF
    n_in+=1
    try:
      aid=int(csvrow['AID'])
      sid=int(csvrow['SID'])
      cid=int(csvrow['CID'])
      outcome=csvrow['Bioactivity Outcome']
    except:
      logging.error('[CID=%d] bad CSV line; problem parsing.'%cid)
      logging.debug('csvrow = %s'%str(csvrow))
      continue
    if cid_query!=cid:
      logging.error('Aaack! [CID=%d] CID mismatch: != %d'%(cid,int(csvrow['CID'])))
      return True
    if aid not in aidset:
      continue
      logging.debug('AID [%d] ok (pre-selected).'%(aid))

    if not csvrow['Activity Value [uM]']:
      continue

    try:
      actval=float(csvrow['Activity Value [uM]'])
    except:
      logging.error('[CID=%d] bad CSV line; problem parsing activity: "%s"'%(cid,csvrow['Activity Value [uM]']))
      continue

    n_act+=1
    outcome_code = OUTCOME_CODES[outcome.lower()] if outcome.lower() in OUTCOME_CODES else 0

    fout.write('%d,%d,%d,%d,%.3f\n'%(cid,sid,aid,outcome_code,actval))

  logging.info('[CID=%d] records: %2d ; activities: %2d'%(cid_query,n_in,n_act))
  return True

#############################################################################



#!/usr/bin/env python3
##############################################################################
### hugo_utils.py - utility functions for HUGO REST API.
### See: http://www.genenames.org/help/rest-web-service-help
##############################################################################
import sys,os,re,json,time,logging

from ..util import rest_utils
#
OFMTS={'XML':'application/xml','JSON':'application/json'}
#
HEADERS={'Accept':OFMTS['JSON']}
#
GENE_TAGS=[
	"alias_symbol",
	"ena",
	"ensembl_gene_id",
	"entrez_id",
	"gene_family",
	"hgnc_id",
	"location",
	"locus_group",
	"locus_type",
	"name",
	"prev_name",
	"prev_symbol",
	"refseq_accession",
	"status",
	"symbol",
	"ucsc_id",
	"uniprot_ids",
	"uuid",
	"vega_id"
	]
#
##############################################################################
def GetGene(qry, ftype, base_url):
  rval=None
  try:
    rval=rest_utils.GetURL(base_url+'/fetch/%s/%s'%(ftype.lower(), qry), headers=HEADERS, parse_json=True)
  except Exception as e:
    logging.error('%s'%(e))
  if type(rval) is not dict: return None
  return rval

##############################################################################
def GetGenes(qrys, ftypes, base_url, fout, found_only, skip):
  '''The API can return multiple hits for each query, though should not 
for exact SYMBOL fetch.  One case is for RPS15P5, status "Entry Withdrawn".'''
  n_in=0;	#queries
  n_found=0;	#queries found (1+ genes)
  n_notfound=0;	#queries not found
  n_ambig=0;	#ambiguous queries
  n_gene=0;	#total gene records returned
  logging.debug('ftypes[%d]: %s'%(len(ftypes), (','.join(ftypes))))
  fout.write(','.join(['query', 'query_type']+GENE_TAGS)+'\n')
  for qry in qrys:
    n_in+=1
    if n_in<skip: continue
    logging.debug('%d. query: %s'%(n_in, qry))
    numFound=0; genes=[];
    ftype_hit=None
    for ftype in ftypes:
      rval = GetGene(qry, ftype, base_url)
      ftype_hit=ftype

      try:
        responseHeader = rval['responseHeader']
        logging.debug('response status = %d'%(responseHeader['status']))
        response = rval['response'] 
        numFound = response['numFound'] if 'numFound' in response else 0
        genes = response['docs']
      except Exception as e:
        logging.error('%s'%(e))

      if numFound>0:
        break

    if numFound==0:
      n_notfound+=1
      logging.debug('Not found: %s = %s'%(ftype_hit, qry))
      if not found_only:
        fout.write(','.join(map(lambda s:'"%s"'%s, [qry, ('|'.join(ftypes))]+['' for j in range(len(GENE_TAGS))]))+'\n')
      continue
    elif numFound>0:
      n_found+=1

    if len(genes)>1:
      logging.debug('Warning: multiple (%d) hits, %s = %s (Duplicate may be status=withdrawn.)'%(len(genes), ftype_hit, qry))
      n_ambig+=1

    for gene in genes:
      vals=[qry, ftype_hit]
      for tag in GENE_TAGS:
        val = gene[tag] if tag in gene else ''
        if type(val) is list: val = ';'.join(val)
        vals.append(val)
      fout.write(','.join(map(lambda s:'"%s"'%s, vals))+'\n')
      n_gene+=1

  logging.info('queries: %d ; found: %d ; not found: %d ; ambiguous: %d ; total genes returned: %d'%(n_in-skip, n_found, n_notfound, n_ambig, n_gene))
  return n_in-skip, n_found

##############################################################################
def SearchGenes(qrys, ftype, base_url, fout):
  n_in=0; n_found=0;
  for qry in qrys:
    n_in+=1
    logging.debug('%d. query: %s'%(n_in, qry))
    url = (base_url+'/search%s/*%s*'%((('/'+ftype.lower()) if ftype else ''), qry))
    try:
      rval=rest_utils.GetURL(url, headers=HEADERS, parse_json=True)
    except Exception as e:
      logging.error('%s'%(e))

    responseHeader = rval['responseHeader']
    logging.debug('response status = %d'%(responseHeader['status']))
    response = rval['response'] 
    numFound = response['numFound'] if 'numFound' in response else 0
    if numFound>0:
      n_found+=1
    fout.write(json.dumps(response, sort_keys=True, indent=2)+'\n')

  return n_in,n_found

##############################################################################
def SearchableFieldsList(base_url):
  try:
    rval=rest_utils.GetURL(base_url+'/info', headers=HEADERS, parse_json=True)
  except Exception as e:
    logging.error('%s'%(e))
  fields = rval['searchableFields'] if 'searchableFields' in rval else None
  fields.sort()
  return fields




#!/usr/bin/env python3
"""
DrugCentral db utility functions.
"""
import os,sys,re,logging
import psycopg2,psycopg2.extras

#############################################################################
def Connect(dbhost, dbport, dbname, dbusr, dbpw):
  """Connect to db; specify default cursor type DictCursor."""
  dsn = ("host='%s' port='%s' dbname='%s' user='%s' password='%s'"%(dbhost, dbport, dbname, dbusr, dbpw))
  dbcon = psycopg2.connect(dsn)
  dbcon.cursor_factory = psycopg2.extras.DictCursor
  return dbcon

#############################################################################
def Version(dbcon, dbschema, fout):
  cur = dbcon.cursor()
  cur.execute("SELECT * FROM {0}.dbversion".format(dbschema))
  row = cur.fetchone()
  tags = list(row.keys())
  fout.write("\t".join(tags)+"\n")
  fout.write("\t".join([str(row[tag]) for tag in tags])+"\n")

#############################################################################
def Describe(dbcon, dbschema, fout):
  '''Describing the schema.'''
  cur = dbcon.cursor()
  cur.execute(("SELECT table_name FROM information_schema.tables WHERE table_schema = %s"), (dbschema,))
  fout.write("schema\ttable\tcolumn\tdatatype\n")
  cur2 = dbcon.cursor()
  for row in cur:
    cur2.execute(("SELECT column_name,data_type FROM information_schema.columns WHERE table_schema = %s AND table_name = %s"), (dbschema, row[0]))
    for row2 in cur2:
      fout.write("\t".join([dbschema, row[0], row2[0], row2[1]])+"\n")
  cur2.close()
  cur.close()

#############################################################################
def Counts(dbcon, dbschema, fout):
  '''Listing the table rowcounts.'''
  n_table=0; n_row=0;
  cur = dbcon.cursor()
  cur.execute(("SELECT table_name FROM information_schema.tables WHERE table_schema = %s ORDER BY table_name"), (dbschema,))
  fout.write("schema\ttable\trowcount\n")
  cur2 = dbcon.cursor()
  for row in cur:
    n_table+=1
    cur2.execute("SELECT count(*) FROM {0}.{1}".format(dbschema, row[0]))
    row2 = cur2.fetchone()
    n_row+=row2[0]
    fout.write("\t".join([dbschema, row[0], str(row2[0])])+"\n")
  cur2.close()
  cur.close()
  logging.info("Totals: tables: {0}; rows: {1}".format(n_table, n_row))

#############################################################################
def ListStructures(dbcon, dbschema, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute("SELECT * FROM {0}.structures".format(dbschema))
  for row in cur:
    if not tags:
      tags = list(row.keys())
      tags.remove("molfile")
      tags.remove("molimg")
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListStructures2Smiles(dbcon, dbschema, fout):
  n_out=0; 
  cur = dbcon.cursor()
  cur.execute("SELECT smiles, id, name FROM {0}.structures WHERE smiles IS NOT NULL".format(dbschema))
  for row in cur:
    vals = [row["smiles"], str(row["id"]), row["name"]]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListStructures2Molfile(dbcon, dbschema, fout):
  n_out=0;
  cur = dbcon.cursor()
  cur.execute("SELECT molfile, id, name FROM {0}.structures WHERE molfile IS NOT NULL".format(dbschema))
  for row in cur:
    molfile = re.sub(r'^[\s]*\n', str(row["id"])+"\n", row["molfile"])
    fout.write(molfile)
    fout.write("> <DRUGCENTRAL_STRUCT_ID>\n"+str(row["id"])+"\n\n")
    fout.write("> <NAME>\n"+row["name"]+"\n\n")
    fout.write("$$$$\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListProducts(dbcon, dbschema, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute("SELECT * FROM {0}.product".format(dbschema))
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListActiveIngredients(dbcon, dbschema, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  cur.execute("SELECT * FROM {0}.active_ingredient".format(dbschema))
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def ListIndications(dbcon, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	omop.umls_cui,
	omop.cui_semantic_type umls_semantic_type,
	omop.snomed_conceptid,
	omop.snomed_full_name
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
"""
  cur.execute(sql)
  for row in cur:
    if not tags:
      tags = list(row.keys())
      fout.write("\t".join(tags)+"\n")
    vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
    fout.write("\t".join(vals)+"\n")
    n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def SearchIndications(dbcon, terms, fout):
  """Search names via Pg regular expression (SIMILAR TO)."""
  n_out=0; tags=None;
  cur = dbcon.cursor()
  for term in terms:
    sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	omop.umls_cui,
	omop.cui_semantic_type umls_semantic_type,
	omop.snomed_conceptid,
	omop.snomed_full_name
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
	AND (omop.concept_name ~* '{0}' OR omop.snomed_full_name ~* '{0}')
""".format(term)
    cur.execute(sql)
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetIndicationStructures(dbcon, ids, fout):
  """Input OMOP conceptIds."""
  n_out=0; tags=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	omop.concept_id omop_concept_id,
	omop.concept_name omop_concept_name,
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	s.inchi,
	s.cd_formula,
	s.cd_molweight
FROM
	omop_relationship omop
JOIN
	structures s ON omop.struct_id = s.id
WHERE
	omop.relationship_name = 'indication'
	AND omop.concept_id = %s
"""
  for id_this in ids:
    cur.execute(sql, (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetStructure(dbcon, ids, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  for id_this in ids:
    cur.execute(("SELECT * FROM structures WHERE id = %s"), (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        tags.remove("molfile")
        tags.remove("molimg")
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetStructureBySynonym(dbcon, ids, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  for id_this in ids:
    cur.execute(("SELECT str.id, str.name structure_name, syn.name synonym FROM structures AS str JOIN synonyms AS syn ON syn.id=str.id WHERE syn.name = %s"), (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetStructureIds(dbcon, ids, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  sql="""\
SELECT
	struct_id,
	id_type,
	identifier AS id
FROM
	identifier
WHERE 
	struct_id = %s
"""
  for id_this in ids:
    cur.execute(sql, (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetStructureProducts(dbcon, ids, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	p.id product_id,
	p.ndc_product_code,
	p.form product_form,
	p.generic_name product_generic_name,
	p.product_name,
	p.route product_route,
	p.marketing_status product_marketing_status,
	p.active_ingredient_count product_active_ingredient_count
FROM
	structures AS s
JOIN 
	active_ingredient ai ON ai.struct_id = s.id
JOIN 
	product p ON p.ndc_product_code = ai.ndc_product_code
WHERE 
	s.id = %s
"""
  for id_this in ids:
    cur.execute(sql, (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################
def GetProductStructures(dbcon, ids, fout):
  n_out=0; tags=None;
  cur = dbcon.cursor()
  sql="""\
SELECT DISTINCT
	p.id product_id,
	p.product_name,
	s.id struct_id,
	s.name struct_name,
	s.smiles,
	s.inchikey,
	s.inchi,
	s.cd_formula,
	s.cd_molweight
FROM
	product p
JOIN 
	active_ingredient ai ON p.ndc_product_code = ai.ndc_product_code
JOIN 
	structures s ON ai.struct_id = s.id
WHERE 
	p.id = %s
"""
  for id_this in ids:
    cur.execute(sql, (id_this,))
    for row in cur:
      if not tags:
        tags = list(row.keys())
        fout.write("\t".join(tags)+"\n")
      vals = [("{:.3f}".format(row[tag]) if type(row[tag]) is float else '' if row[tag] is None else str(row[tag])) for tag in tags]
      fout.write("\t".join(vals)+"\n")
      n_out+=1
  logging.info("n_out: {0}".format(n_out))

#############################################################################



#!/usr/bin/env python3
"""
GeneOntology client.
"""
###
import os,sys,argparse,re,json,time,logging
import urllib.parse

from .. import geneontology

API_HOST='api.geneontology.org'
API_BASE_PATH='/api'
#
##############################################################################
if __name__=='__main__':
  epilog="""\
"""
  parser = argparse.ArgumentParser(description='GeneOntolgy API client', epilog=epilog)
  ops=['list_Terms', 'list_Genes', 'get_Entities', 'get_GeneTerms' ]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input file of IDs")
  parser.add_argument("--ids", help="ID list (comma-separated)(e.g. NCBIGene:84570, GO:0006954)")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL = 'https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout = open(args.ofile, "w+")
  else:
    fout = sys.stdout

  if args.ids:
    ids = re.split(r'[,\s]+', args.ids.strip())
  elif args.ifile:
    with open(args.ifile) as fin:
      while True:
        line = fin.readline()
        if not line: break
        ids.append(line.rstrip())

  if args.op == 'list_Terms':
    geneontology.Utils.ListTerms(BASE_URL, fout)

  elif args.op == 'list_Genes':
    geneontology.Utils.ListGenes(BASE_URL, fout)

  elif args.op == 'get_Entities':
    geneontology.Utils.GetEntities(BASE_URL, ids, fout)

  elif args.op == 'get_GeneTerms':
    geneontology.Utils.GetGeneTerms(BASE_URL, ids, fout)

  else:
    parser.error('Invalid operation: %s'%args.op)




#!/usr/bin/env python3
"""
See: http://www.genenames.org/
"""
###
import sys,os,re,argparse,time,json,logging
#
from ..util import rest_utils
from .. import hugo
#
API_HOST='rest.genenames.org'
API_BASE_PATH=''
#
##############################################################################
if __name__=='__main__':
  FTYPES=['SYMBOL','ALIAS_SYMBOL', 'HGNC_ID', 'UNIPROT', 'ENTREZ_ID', 'ENSEMBL_GENE_ID', 'VEGA_ID'];
  epilog='''\
           SYMBOL: HGNC gene symbol (e.g. ZNF3),
     ALIAS_SYMBOL: synonym gene symbol (e.g. AAA1),
          HGNC_ID: HGNC gene ID (e.g. HGNC:13089),
          UNIPROT: UniProt ID,
        ENTREZ_ID: NCBI Gene ID,
  ENSEMBL_GENE_ID: Ensembl Gene ID (e.g. ENSG00000003056),
          VEGA_ID:
'''
  parser = argparse.ArgumentParser(description='HUGO HGNC REST API client', epilog=epilog)
  ops = ['search', 'get', 'show_info', 'list_searchable']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="gene query")
  parser.add_argument("--qfile", help="query file")
  parser.add_argument("--ftypes", default="SYMBOL,ALIAS_SYMBOL", help="comma-separated list")
  parser.add_argument("--ftypes_all", type=bool, help="(could be slow)")
  parser.add_argument("--found_only", type=bool, help="no output for not-found")
  parser.add_argument("--rawquery")
  parser.add_argument("--nmax", type=int, default=None, help="max records")
  parser.add_argument("--skip", type=int, default=0, help="skip 1st SKIP queries")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  API_BASE_URL='https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout=open(args.ofile,"w")
    if not fout: parser.error('ERROR: cannot open outfile: %s'%args.ofile)
  else:
    fout=sys.stdout

  t0=time.time()

  qrys=[];
  if args.qfile:
    fin=open(args.qfile)
    if not fin: parser.error('ERROR: cannot open qfile: %s'%args.qfile)
    while True:
      line=fin.readline()
      if not line: break
      if line.rstrip(): qrys.append(line.rstrip())
    logging.info('input queries: %d'%(len(qrys)))
    fin.close()
  elif args.query:
    qrys.append(args.query)

  if args.ftypes_all:
    ftypes=hugo.Utils.SearchableFieldsList(API_BASE_URL)
  else:
    ftypes = re.split('[, ]+', args.ftypes.strip())
    ftypes = [s.upper() for s in ftypes]
    if len(set(ftypes) - set(FTYPES))>0:
      parser.error("Invalid field type[s]: {0}".format(set(ftypes) - set(FTYPES)))

  if args.op == "show_info":
    rval=rest_utils.GetURL(API_BASE_URL+'/info',headers={'Accept':hugo.Utils.OFMTS['JSON']},parse_json=True)
    print(json.dumps(rval,sort_keys=True,indent=2))

  elif args.op == "list_searchable":
    fields=hugo.Utils.SearchableFieldsList(API_BASE_URL)
    for i,field in enumerate(fields):
      print('\t%d. %s'%(i+1,field))

  elif args.op == "get":
    fields=hugo.Utils.SearchableFieldsList(API_BASE_URL)
    for ftype in ftypes:
      if ftype.lower() not in fields:
        parser.error("ftype \"%s\" not searchable field."%ftype)
    n_in,n_hit = hugo.Utils.GetGenes(qrys, ftypes, API_BASE_URL, fout, args.found_only, args.skip)
    logging.info('queries: %d, found: %d, not found: %d'%(n_in, n_hit, n_in-n_hit))

  elif args.op == "search":
    fields=hugo.Utils.SearchableFieldsList(API_BASE_URL)
    if ftype.lower() not in fields:
      parser.error("ftype \"%s\" not searchable field."%ftype)
    rval=hugo.Utils.SearchGenes(qrys, ftype, API_BASE_URL, fout)

  else:
    parser.error("Invalid operation: %s"%args.op)

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
"""
Access to Ensembl REST API.
http://rest.ensembl.org/documentation/info/lookup
"""
import sys,os,re,time,json,logging
import requests
#
##############################################################################
def ShowVersion(base_url, fout):
  rest_api_ver = requests.get(base_url+'/info/rest?content-type=application/json').json()['release']
  sw_api_ver = requests.get(base_url+'/info/software?content-type=application/json').json()['release']
  eg_ver = requests.get(base_url+'/info/eg_version?content-type=application/json').json()['version']
  fout.write("EnsEMBL REST API version: {0}; EnsEMBL software API version: {1}; EnsEMBL genomes version: {2}\n".format(rest_api_ver, sw_api_ver, eg_ver))

##############################################################################
def ListSpecies(base_url, fout):
  n_out=0; tags=None;
  rval = requests.get(base_url+'/info/species?content-type=application/json').json()
  specs = rval["species"]
  for spec in specs:
    if not tags:
      tags = spec.keys()
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(spec[tag]) if tag in spec else "") for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('n_out: {0}'.format(n_out))

##############################################################################
def GetInfo(base_url, ids, fout):
  n_gene=0; n_out=0; n_err=0; tags=[];
  for id_this in ids:
    url_this = base_url+'/lookup/id/'+id_this+'?content-type=application/json&expand=0'
    logging.debug(url_this)
    rval = requests.get(url_this, headers={"Content-Type":"application/json"})
    if not rval.ok:
      logging.error('{0} : "{1}"'.format(rval.status_code, id_this))
      n_err+=1
      continue
    gene = rval.json()
    if not tags:
      for tag in gene.keys():
        if type(gene[tag]) not in (list, dict): tags.append(tag) #Only simple metadata.
      fout.write('\t'.join(tags)+'\n')
    n_gene+=1
    vals = [str(gene[tag]).replace('\n', ' ') if tag in gene and gene[tag] is not None else '' for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('n_ids: {0}; n_gene: {1}; n_out: {2}; n_err: {3}'.format(len(ids), n_gene, n_out, n_err))

##############################################################################
def GetXrefs(base_url, ids, fout):
  n_out=0; n_err=0; tags=None; dbcounts={};
  for id_this in ids:
    url_this = base_url+'/xrefs/id/'+id_this
    rval = requests.get(url_this, headers={"Content-Type":"application/json"})
    if not rval.ok:
      #rval.raise_for_status()
      logging.error('{0} : "{1}"'.format(rval.status_code, id_this))
      n_err+=1
      continue
    xrefs = rval.json()
    for xref in xrefs:
      if not (type(xref) is dict and 'dbname' in xref): continue
      dbname = xref['dbname']
      if dbname not in dbcounts: dbcounts[dbname]=0
      dbcounts[dbname]+=1
      if not tags:
        tags = list(xref.keys())
        fout.write('\t'.join(["id"]+tags)+'\n')
      vals = [(str(xref[tag]) if tag in xref else '') for tag in tags]
      fout.write('\t'.join([id_this]+vals)+'\n')
      n_out+=1
  for key in sorted(dbcounts.keys()):
    logging.info('Xref counts, db = %12s: %5d'%(key, dbcounts[key]))
  logging.info('n_ids: {0}; n_out: {1}; n_err: {2}'.format(len(ids), n_out, n_err))

##############################################################################



#!/usr/bin/env python3
"""
GWAS Catalog REST API client.

https://www.ebi.ac.uk/gwas/docs/api
https://www.ebi.ac.uk/gwas/rest/api
https://www.ebi.ac.uk/gwas/rest/docs/api
https://www.ebi.ac.uk/gwas/rest/docs/sample-scripts
"""
import sys,os,re,argparse,json,time,logging
#
from .. import gwascatalog
#
API_HOST='www.ebi.ac.uk'
API_BASE_PATH='/gwas/rest/api'
#
##############################################################################
if __name__=='__main__':
  epilog = "Examples: PubmedId=28530673; gcst=GCST004364,GCST000227; EfoUri=EFO_0004232; snpId=rs6085920"
  parser = argparse.ArgumentParser(description='GWAS Catalog REST API client', epilog=epilog)
  searchtypes=['pubmedmid', 'gcst', 'efotrait', 'efouri', 'accessionid', 'rs']
  ops = ['list_studies', 'search_studies', 'get_studyAssociations', 'get_snps']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", dest="ids", help="IDs, comma-separated")
  parser.add_argument("--searchtype", choices=searchtypes, help="ID type")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, 'w') if args.ofile else sys.stdout

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  t0 = time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split(r'\s*,\s*', args.ids.strip())

  if args.op == 'search_studies':
    gwascatalog.Utils.SearchStudies(base_url, ids, args.searchtype, fout)

  elif args.op == 'list_studies':
    gwascatalog.Utils.ListStudies(base_url, fout)

  elif args.op == 'get_studyAssociations':
    gwascatalog.Utils.GetStudyAssociations(base_url, ids, fout)

  elif args.op == 'get_snps':
    gwascatalog.Utils.GetSnps(base_url, ids, fout)

  else:
    parser.error("Unknown operation: {0}".format(args.op))

  logging.info('{0}: elapsed time: {1}'.format(os.path.basename(sys.argv[0]), time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python3
"""
MeSH XML utility functions.

MeSH XML
Download: https://www.nlm.nih.gov/mesh/download_mesh.html
Doc: https://www.nlm.nih.gov/mesh/xml_data_elements.html
"""
###
import sys,os,re,logging
import xml.etree.cElementTree as ElementTree
from xml.parsers import expat

#############################################################################
def Desc2Csv(branch, fin, fout):
  fout.write('id\ttreenum\tterm\n')
  n_elem=0; n_term=0;
  for event, elem in ElementTree.iterparse(fin):
    meshid,meshterm,desclass,treenum = None,None,None,None;
    n_elem+=1
    logging.debug('Event:%6s, elem.tag:%6s, elem.text:"%6s"'%(event,elem.tag,elem.text.strip() if elem.text else ''))
    if elem.tag == 'DescriptorRecord':
      if elem.attrib['DescriptorClass'] != '1': continue
      meshid,meshterm,desclass,treenum = None,None,None,None;
      meshid=elem.findtext('DescriptorUI')
      #treenum=elem.findtext('TreeNumberList/TreeNumber')
      elems = elem.findall('TreeNumberList/TreeNumber')
      treenums = map(lambda e: e.text, elems)
      for treenum in treenums:
        if (re.match(r'^%s'%branch,treenum)): break
      meshterm=elem.findtext('DescriptorName/String')
      #See also: ConceptList/Concept/TermList/Term/String (may be multiple)

      if not meshid:
        logging.info('skipping, no meshid: %s'%str(elem))
      elif not meshterm:
        logging.info('meshid: %s ; skipping, no meshterm'%meshid)
      elif not treenum:
        logging.info('meshid: %s ; skipping, no treenum'%meshid)
      elif not re.match(r'^%s'%branch,treenum):
        logging.info('meshid: %s ; skipping, non-%s treenum: %s'%(meshid,branch,treenum))
      else:
        fout.write('%s\t%s\t%s\n'%(meshid,treenum,meshterm))
        n_term+=1

  logging.info('DEBUG: n_elem: %d'%n_elem)
  logging.info('DEBUG: n_term: %d'%n_term)

#############################################################################
### SCRClass
### Description: Attribute of <DescriptorRecord> one of:
### 1 = Regular chemical, drug, or other substance (the most common type)
### 2 = Protocol, for example, FOLFOXIRI protocol
### 3 = Rare disease, for example, Canicola fever
### Subelement of: not applicable. Attribute of <SupplementalRecord>
### Record Type: SCR
### https://www.nlm.nih.gov/mesh/xml_data_elements.html
#############################################################################
### There are non-disease records mapped to diseases, e.g. C041229,
### "GHM protein, human", SCRClass=1, is mapped to D006362, "Heavy Chain Disease".
### These cannot be identified from the supplementary file alone.
#############################################################################
def Supp2Csv(branch, fin, fout):
  fout.write('id\tterm\tid_to\tterm_to\n')
  n_elem=0; n_term=0;
  for event,elem in ElementTree.iterparse(fin):
    meshid,name,meshid_to,meshterm_to = None,None,None,None;
    n_elem+=1
    logging.debug('Event:%6s, elem.tag:%6s, elem.text:"%6s"'%(event,elem.tag,elem.text.strip() if elem.text else ''))
    if elem.tag == 'SupplementalRecord':
      scrclass = elem.attrib['SCRClass']
      if branch not in ('C', 'D'): continue
      if branch=='C' and scrclass!='3': continue #disease-only
      if branch=='D' and scrclass!='1': continue #chemical-only
      meshid=elem.findtext('SupplementalRecordUI')
      name=elem.findtext('SupplementalRecordName/String')
      mtlist=elem.find('HeadingMappedToList')
      if mtlist is None:
        continue
      meshid_to=mtlist.findtext('HeadingMappedTo/DescriptorReferredTo/DescriptorUI')
      meshterm_to=mtlist.findtext('HeadingMappedTo/DescriptorReferredTo/DescriptorName/String')
      fout.write('%s\t%s\t%s\t%s\n'%(meshid,name,meshid_to,meshterm_to))
      n_term+=1
  logging.debug('n_elem: %d'%n_elem)
  logging.info('n_term: %d'%n_term)

#############################################################################



#!/usr/bin/env python3
"""
Utility for STRING Db REST API.

STRING = Search Tool for the Retrieval of Interacting Genes/Proteins

See: http://string-db.org/help/api/

http://[database]/[access]/[format]/[request]?[parameter]=[value]

database:
string-db.org
string.embl.de
stitch.embl.de
"""
###
import sys,os,re,json,argparse,time,logging
import urllib,urllib.request,urllib.parse
#
from ..util import rest_utils
#
API_HOST='string-db.org'
API_BASE_PATH='/api'
#
NETWORK_FLAVORS=['evidence','confidence','actions']
IMG_FMTS=['image','highres_image','svg']
#
##############################################################################
def GetIds(base_url, pids, fout):
  url=base_url+'/tsv/get_string_ids'
  if len(pids)==1:
    url+=('?identifier=%s'%pids[0])
  else:
    url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('queries: %d ; results: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetInteractionPartners(base_url, pids, species, minscore, fout):
  url=base_url+'/tsv/interaction_partners'
  if len(pids)==1:
    url+=('?identifier=%s'%pids[0])
  else:
    url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('queries: %d ; interaction partners: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetEnrichment(base_url, pids, species, minscore, fout):
  url=base_url+'/tsv/enrichment'
  url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('query genes: %d ; enrichment results: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetPPIEnrichment(base_url, pids, species, minscore, fout):
  url=base_url+'/tsv/ppi_enrichment'
  url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('query genes: %d ; enrichment results: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetNetwork(base_url, pid, species, minscore, netflavor, fout):
  """Format: PNG image."""
  url=base_url+'/tsv/network?identifier=%s'%(pid)
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  url+=('&network_flavor=%s'%netflavor)
  rval=rest_utils.GetURL(url, parse_json=False, parse_xml=False)
  fout.write(rval)
  logging.info('edges: %d'%(len(rval.splitlines())-1))

##############################################################################
def GetNetworkImage(base_url, pid, species, minscore, netflavor, imgfmt, fout):
  url=base_url+'/%s/network?identifier=%s'%(imgfmt, pid)
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  url+=('&network_flavor=%s'%netflavor)
  rval=rest_utils.GetURL(url, parse_json=False, parse_xml=False)
  fout.write(rval)

##############################################################################
def GetInteractors(base_url, pids, species, minscore, fout):
  """2018: May be deprecated since 2015."""
  url=base_url+'/tsv/interactors'
  if len(pids)==1:
    url+=('?identifier=%s'%pids[0])
  else:
    url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('queries: %d ; interactors: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetActions(base_url, pids, species, minscore, fout):
  """2018: May be deprecated since 2015."""
  url=base_url+'/tsv/actions'
  if len(pids)==1:
    url+=('?identifier=%s'%pids[0])
  else:
    url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  url+=('&required_score=%d'%minscore)
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('queries: %d ; actions: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
def GetAbstracts(base_url, pids, species, fout):
  """2018: May be deprecated since 2015."""
  url=base_url+'/tsv/abstracts'
  if len(pids)==1:
    url+=('?identifier=%s'%pids[0])
  else:
    url+=('?identifiers=%s'%urllib.parse.quote('\n'.join(pids)))
  url+=('&species=%s'%species if species else '')
  rval=rest_utils.GetURL(url, parse_xml=False, parse_json=False)
  fout.write(rval)
  logging.info('queries: %d ; abstracts: %d'%(len(pids), len(rval.splitlines())-1))

##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(
        description='STRING-DB REST API client utility',
        epilog="""\
Example protein IDs: DRD1 DRD1_HUMAN DRD2 DRD2_HUMAN ;
Example species: 9606 (human, via taxon identifiers, http://www.uniprot.org/taxonomy) ;
Image formats: PNG PNG_highres SVG ;
MAY BE DEPRECATED: getInteractors, getActions, getAbstracts
""")
  ops = ['getIds', 'getInteractionPartners',
	'getNetwork', 'getNetworkImage',
	'getEnrichment', 'getPPIEnrichment',
	'getInteractors', 'getActions', 'getAbstracts']
  parser.add_argument("op",choices=ops,help='operation')
  parser.add_argument("--id", dest="id", help="protein ID (ex:DRD1_HUMAN)")
  parser.add_argument("--ids", dest="ids", help="protein IDs, comma-separated")
  parser.add_argument("--idfile", dest="idfile", help="input file, protein IDs")
  parser.add_argument("--o", dest="ofile", help="output file")
  parser.add_argument("--species", help="taxon code, ex: 9606 (human)")
  parser.add_argument("--minscore", type=int, default=500, help="signifcance threshold 0-1000")
  parser.add_argument("--netflavor", choices=NETWORK_FLAVORS, default='evidence', help="network flavor")
  parser.add_argument("--imgfmt", choices=IMG_FMTS, default='image', help="image format")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  pids=[]
  if args.idfile:
    fin=open(args.idfile)
    if not fin: parser.error('failed to open input file: %s'%args.idfile)
    while True:
      line=fin.readline()
      if not line: break
      pids.append(line.strip())
  elif args.ids:
    pids = re.split(r'\s*,\s*', args.ids.strip())
  elif args.id:
    pids=[args.id]

  if args.ofile:
    fout=open(args.ofile, "w+")
    if not fout: parser.error('failed to open output file: %s'%args.ofile)
  else:
    fout=sys.stdout

  t0=time.time()

  if args.op == 'getIds':
    if not pids: parser.error('PID[s] required.')
    GetIds(base_url, pids, fout)

  elif args.op == 'getInteractors':
    if not pids: parser.error('PID[s] required.')
    GetInteractors(base_url, pids, args.species, args.minscore, fout)

  elif args.op == 'getInteractionPartners':
    if not pids: parser.error('PID[s] required.')
    GetInteractionPartners(base_url, pids, args.species, args.minscore, fout)

  elif args.op == 'getActions':
    if not pids: parser.error('PID[s] required.')
    GetActions(base_url, pids, args.species, args.minscore, fout)

  elif args.op == 'getEnrichment':
    if not len(pids)>1: parser.error('PIDs (2+) required.')
    GetEnrichment(base_url, pids, args.species, args.minscore, fout)

  elif args.op == 'getPPIEnrichment':
    if not len(pids)>1: parser.error('PIDs (2+) required.')
    GetPPIEnrichment(base_url, pids, args.species, args.minscore, fout)

  elif args.op == 'getAbstracts':
    if not pids: parser.error('PID[s] required.')
    GetAbstracts(base_url, pids, args.species, fout)

  elif args.op == 'getNetwork':
    if not pids: parser.error('PID required.')
    GetNetwork(base_url, pids[0], args.species, args.minscore, args.netflavor, fout)

  elif args.op == 'getNetworkImage':
    if not pids: parser.error('PID required.')
    if not args.ofile: parser.error('--o OUTFILE required.')
    fout.close()
    fout = open(args.ofile, "wb+") #binary
    if not fout: parser.error('failed to open output file: %s'%args.ofile)
    GetNetworkImage(base_url, pids[0], args.species, args.minscore, args.netflavor, args.imgfmt, fout)

  else:
    parser.print_help()

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))




#!/usr/bin/env python3
"""
TCRD db client utility (see also REST API client)
In general, return data as lists and dicts, readily convertible to JSON.
See https://dev.mysql.com/doc/connector-python/en/
regarding mysql-connector-python package.
"""
import os,sys,argparse,re,time,json,logging
import mysql.connector as mysql

#############################################################################
def ListTables(dbcon):
  cur=dbcon.cursor()
  cur.execute('SHOW TABLES')
  rows=cur.fetchall()
  tables=[]
  for row in rows:
    tables.append(row[0])
  tables.sort()
  return tables

#############################################################################
def Info(dbcon):
  cur=dbcon.cursor(dictionary=True)
  cur.execute('SELECT * FROM dbinfo')
  row=cur.fetchone()
  return row

#############################################################################
def Describe(dbcon):
  cur=dbcon.cursor()
  outtxt=''
  for table in ListTables(dbcon):
    outtxt+=('%s:\n'%table)
    cur.execute('DESCRIBE '+table)
    row=cur.fetchone()
    cols=[]
    while row:
      cols.append(row[0])
      row=cur.fetchone()
    cols.sort()
    outtxt+=('\t'+(', '.join(cols))+'\n')
  return outtxt

#############################################################################
def Counts(dbcon):
  tables=ListTables(dbcon)
  cur=dbcon.cursor()
  outtxt=''
  for table in tables:
    cur.execute('SELECT count(*) FROM '+table)
    row=cur.fetchone()
    outtxt+=('%18s: %8d rows\n'%(table,row[0]))
  return outtxt

#############################################################################
def AttributeCounts(dbcon):
  cur=dbcon.cursor()
  sql='SELECT ga.type, COUNT(ga.id) FROM gene_attribute ga GROUP BY ga.type'
  cur.execute(sql)
  row=cur.fetchone()
  attrcounts={}
  while row:
    attrcounts[row[0]]=row[1]
    row=cur.fetchone()

  outtxt='Attribute counts:\n'
  for attr_type in sorted(attrcounts.keys()):
    outtxt+=('%13s: %8d\n'%(attr_type,attrcounts[attr_type]))
  return outtxt

#############################################################################
def XrefCounts(dbcon):
  '''Synonyms, aliases, and Xrefs.'''
  cur=dbcon.cursor()
  outtxt=''

  f,t = 'name','synonym'
  cur.execute('SELECT COUNT(DISTINCT %s) FROM %s'%(f,t))
  row=cur.fetchone()
  outtxt+=('%s_count: %8d\n'%(t,row[0]))

  f,t = 'name','alias'
  cur.execute('SELECT COUNT(DISTINCT %s) FROM %s'%(f,t))
  row=cur.fetchone()
  outtxt+=('%s_count: %8d\n'%(t,row[0]))

  f,t = 'value','xref'
  cur.execute('SELECT COUNT(DISTINCT %s) FROM %s'%(f,t))
  row=cur.fetchone()
  outtxt+=('%s_count: %8d\n'%(t,row[0]))

  cur.execute('SELECT xtype, COUNT(DISTINCT value) FROM xref GROUP BY xtype ORDER BY xtype')
  rows=cur.fetchall()
  for row in rows:
    outtxt+=('%18s: %8d\n'%(row[0],row[1]))

  return outtxt

#############################################################################
def TDLCounts(dbcon):
  cur=dbcon.cursor()
  sql='''\
SELECT tdl,COUNT(id)
FROM target
WHERE tdl in (%(TDLS)s)
GROUP BY tdl
'''%{'TDLS':"'"+("','".join(TDLS))+"'"}
  cur.execute(sql)
  rows=cur.fetchall()
  #outtxt=''
  rdata={}
  n_total=0
  for row in rows:
    #outtxt+=('%8s: %6d\n'%(row[0],row[1]))
    rdata[row[0]] = row[1]
    n_total+=row[1]
  rdata['total'] = n_total
  #outtxt+=('   Total: %6d\n'%(n_total))
  return rdata

#############################################################################
def ListXreftypes(dbcon):
  cur=dbcon.cursor(mysql.cursors.DictCursor)
  sql='SELECT DISTINCT xtype FROM xref ORDER BY xtype'
  cur.execute(sql)
  rows=cur.fetchall()
  xreftypes=[]
  for row in rows:
    if 'xtype' in row: xreftypes.append(row['xtype'])
  return xreftypes

#############################################################################
def ListXrefs(dbcon,qtype,fout):
  cur=dbcon.cursor(mysql.cursors.DictCursor)
  fout.write('"%s","target_id","protein_id"\n'%(qtype))
  cols=['value','target_id','protein_id']
  sql='''\
SELECT DISTINCT %(COLS)s
FROM xref
WHERE xtype = '%(QTYPE)s'
'''%{	'COLS':(','.join(cols)),
	'QTYPE':qtype }
  cur.execute(sql)
  row=cur.fetchone()
  n_xref=0;
  while row:
    tid = row['target_id'] if ('target_id' in row and row['target_id']!=None) else ''
    pid = row['protein_id'] if ('protein_id' in row and row['protein_id']!=None) else ''
    fout.write('"%s",%s,%s\n'%(row['value'],tid,pid))
    n_xref+=1
    row=cur.fetchone()
  logging.info('Xrefs count: %d'%(n_xref))

#############################################################################
def ListTargets(dbcon, tdl, pfam, fout):
  cur = dbcon.cursor(dictionary=True)
  sql='''
SELECT
	target.id AS "tcrdTargetId",
	target.name AS "tcrdTargetName",
	target.fam AS "tcrdTargetFamily",
	target.tdl AS "TDL",
	target.ttype AS "tcrdTargetType",
	target.idg AS "idgList",
	protein.id AS "tcrdProteinId",
	protein.sym AS "tcrdGeneSymbol",
	protein.family AS "tcrdProteinFamily",
	protein.geneid AS "ncbiGeneId",
	protein.uniprot AS "uniprotId",
	protein.up_version AS "uniprotVersion",
	protein.chr,
	protein.description AS "tcrdProteinDescription",
	protein.dtoid AS "dtoId",
	protein.dtoclass AS "dtoClass",
	protein.stringid AS "ensemblProteinId",
	xref.value AS "ensemblGeneId"
FROM
	target
JOIN
	t2tc ON t2tc.target_id = target.id
JOIN
	protein ON protein.id = t2tc.protein_id
JOIN
	xref ON xref.protein_id = protein.id
'''
  wheres=["xref.xtype = 'Ensembl'", "xref.value REGEXP '^ENSG'"]
  if tdl:
    wheres.append('target.tdl = \'%s\''%tdl)
  if pfam:
    wheres.append('target.fam = \'%s\''%pfam)
  if wheres:
    sql+=(' WHERE '+(' AND '.join(wheres)))
  cur.execute(sql)
  row = cur.fetchone()
  i_row=0;
  tids=set(); pids=set(); uniprots=set(); ensps=set(); ensgs=set(); symbs=set();
  while row:
    i_row+=1
    if i_row==1:
      colnames = list(row.keys())
      fout.write('\t'.join(colnames)+'\n')
    if 'tcrdTargetId' in row: tids.add(row['tcrdTargetId'])
    if 'tcrdProteinId' in row: pids.add(row['tcrdProteinId'])
    if 'uniprotId' in row: uniprots.add(row['uniprotId'])
    if 'ensemblProteinId' in row: ensps.add(row['ensemblProteinId'])
    if 'ensemblGeneId' in row: ensgs.add(row['ensemblGeneId'])
    if 'tcrdGeneSymbol' in row: symbs.add(row['tcrdGeneSymbol'])
    fout.write('\t'.join([(str(row[tag]) if tag in row else '') for tag in colnames])+'\n')
    row = cur.fetchone()
  logging.info('TDL: %s'%(tdl if tdl else 'all'))
  logging.info('rows: %d'%(i_row))
  logging.info('Target count: %d'%(len(tids)))
  logging.info('Protein count: %d'%(len(pids)))
  logging.info('Uniprot count: %d'%(len(uniprots)))
  logging.info('ENSP count: %d'%(len(ensps)))
  logging.info('ENSG count: %d'%(len(ensgs)))
  logging.info('GeneSymbol count: %d'%(len(symbs)))

#############################################################################
def GetTargets(dbcon, qs, qtype, fout):
  '''Write data to CSV file, and return as list of dict-rows.'''
  if not qs:
    logging.info('ERROR: no query ID.')
    return
  cur = dbcon.cursor(dictionary=True)
  cols=[
	'target.description',
	'target.id',
	'target.fam',
	'target.name',
	'target.tdl',
	'target.ttype',
	'protein.chr',
	'protein.description',
	'protein.family',
	'protein.geneid',
	'protein.id',
	'protein.name',
	'protein.sym',
	'protein.uniprot'
	]
  if fout: fout.write('query,qtype,'+('%s\n'%(','.join(cols))))

  n_hit=0;
  tids_all=set();
  for qid in qs:
    sql='''\
SELECT
	%(COLS)s
FROM
	target
JOIN
	t2tc ON target.id = t2tc.target_id
JOIN
	protein ON protein.id = t2tc.protein_id
'''%{'COLS':(','.join(map(lambda s: '%s AS %s'%(s, s.replace('.', '_')), cols)))}
    wheres=[]
    if qtype.lower()=='tid':
      wheres.append('target.id = %d'%int(qid))
    elif qtype.lower()=='uniprot':
      wheres.append('protein.uniprot = \'%s\''%qid)
    elif qtype.lower() in ('gid', 'geneid'):
      wheres.append('protein.geneid = \'%s\''%qid)
    elif qtype.lower() in ('genesymb', 'genesymbol'):
      wheres.append('protein.sym = \'%s\''%qid)
    elif qtype.lower() == 'ensp':
      wheres.append('protein.stringid = \'%s\''%qid)
    elif qtype.lower() == 'ncbi_gi':
      sql+=('\nJOIN\n\txref ON xref.protein_id = protein.id')
      wheres.append('xref.xtype = \'NCBI GI\'')
      wheres.append('xref.value = \'%s\''%qid)
      wheres.append('xref.protein_id = protein.id')
    else:
      logging.info('ERROR: unknown query type: %s'%qtype)
      return
    sql+=(' WHERE '+(' AND '.join(wheres)))
    logging.debug('"%s"'%sql)
    logging.debug('query: %s = "%s" ...'%(qtype, qid))
    cur.execute(sql)
    rows=cur.fetchall()
    if not rows:
      logging.info('Not found: %s = %s'%(qtype, qid))
      continue
    n_hit+=1
    tids_this=set();
    colnames=map(lambda s: s.replace('.', '_'), cols)
    for row in rows:
      if 'target_id' in row: tids_this.add(row['target_id'])
      if fout: fout.write('"%s","%s"'%(qid, qtype))
      for j,tag in enumerate(colnames):
        val = row[tag] if tag in row else ''
        if fout: fout.write(',"%s"'%(val))
      if fout: fout.write('\n')
    tids_all |= tids_this

  logging.info('%ss queries: %d, found: %d, not found: %d'%(qtype, len(qs), n_hit, (len(qs)-n_hit)))
  logging.info('Targets found: %d'%(len(tids_all)))
  return list(tids_all)

#############################################################################
def GetPathways(dbcon, tids, fout):
  cur=dbcon.cursor(mysql.cursors.DictCursor)
  cols=[
	't2p.target_id',
	't2p.id',
	't2p.source',
	't2p.id_in_source',
	't2p.name',
	't2p.description',
	't2p.url'
	]
  if fout: fout.write('%s\n'%(','.join(cols)))

  n_hit=0;
  pids_all=set();
  for tid in tids:
    sql='''\
SELECT
	%(COLS)s
FROM
	target2pathway t2p
JOIN
	target t ON t.id = t2p.target_id
WHERE
	t.id = %(TID)s
ORDER BY
	t2p.target_id,
	t2p.id
'''%{	'COLS':(','.join(map(lambda s: '%s AS %s'%(s,s.replace('.','_')),cols))),
	'TID':str(tid)
	}

    logging.debug('"%s"'%sql)
    logging.debug('query: %s ...'%(tid))
    cur.execute(sql)
    rows=cur.fetchall()
    if not rows:
      logging.info('Not found: %s'%(str(tid)))
      continue
    n_hit+=1
    pids_this=set();
    colnames=map(lambda s: s.replace('.','_'),cols)
    for row in rows:
      if 'id' in row: pids_this.add(row['id'])
      for j,tag in enumerate(colnames):
        val = row[tag] if tag in row else ''
        if type(val) is str: val = val.encode('string-escape')
        if fout: fout.write(',"%s"'%(val))
      if fout: fout.write('\n')
    pids_all |= pids_this

  logging.info('queries: %d, found: %d, not found: %d'%(len(tids),n_hit,(len(tids)-n_hit)))
  logging.info('Pathways found: %d'%(len(pids_all)))
  return list(pids_all)

#############################################################################
if __name__=='__main__':
  DBHOST='tcrd.kmc.io'
  DBNAME='tcrd'
  DBUSR='tcrd'
  DBPW=''
  TDLS=['Tdark', 'Tbio', 'Tchem', 'Tclin']
  qtypes=[ 'TID', 'GENEID', 'UNIPROT', 'GENESYMB', 'NCBI_GI', 'ENSP']
  parser = argparse.ArgumentParser(description='TCRD MySql client utility')
  ops = ['info', 'describe', 'counts', 'tdlCounts', 'xrefCounts', 'attrCounts', 'listTargets', 'listXreftypes', 'listXrefs', 'getTargets', 'getTargetpathways']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="query ID or symbol")
  parser.add_argument("--qfile", help="input query file")
  parser.add_argument("--qtype", choices=qtypes, default='TID', help='query type')
  parser.add_argument("--id", help="query ID or symbol")
  parser.add_argument("--tdl", help="Target Development Level (TDL) %s"%('|'.join(TDLS)))
  parser.add_argument("--fam", help="target family GPCR|Kinase|IC|NR|...|Unknown")
  parser.add_argument("--dbname", default=DBNAME)
  parser.add_argument("--dbhost", default=DBHOST)
  parser.add_argument("--dbusr", default=DBUSR)
  parser.add_argument("--dbpw", default=DBPW)
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  qs=[]
  if args.qfile:
    fin = open(args.qfile)
    while True:
      line = fin.readline()
      if not line: break
      qs.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(qs)))
    fin.close()
  elif args.query:
    qs.append(args.query)

  if args.dbusr is not None and args.dbpw is not None:
    dbcon = mysql.connect(host=args.dbhost, user=args.dbusr, passwd=args.dbpw, db=args.dbname)
  else:
    dbcon = mysql.connect(host=args.dbhost, db=args.dbname)

  if args.op=='describe':
    print(Describe(dbcon))

  elif args.op=='info':
    rdata = Info(dbcon)
    print(json.dumps(rdata, indent=2, sort_keys=True))

  elif args.op=='counts':
    print(Counts(dbcon))

  elif args.op=='xrefCounts':
    print(XrefCounts(dbcon))

  elif args.op=='attrCounts':
    print(AttributeCounts(dbcon))

  elif args.op=='tdlCounts':
    rdata = TDLCounts(dbcon)
    print(json.dumps(rdata, indent=2, sort_keys=True))

  elif args.op=='listTargets':
    ListTargets(dbcon, args.tdl, args.fam, fout)

  elif args.op=='getTargets':
    GetTargets(dbcon, qs, args.qtype, fout)

  elif args.op=='getTargetpathways':
    tids = GetTargets(dbcon, qs, args.qtype, None)
    GetPathways(dbcon, tids, fout)

  elif args.op=='listXreftypes':
    xreftypes = ListXreftypes(dbcon)
    print(str(xreftypes))

  elif args.op=='listXrefs':
    xreftypes = ListXreftypes(dbcon)
    if qtype not in xreftypes:
      ErrorExit('ERROR: qtype "%s" invalid.  Available xref types: %s'%(qtype, str(xreftypes)))
    ListXrefs(dbcon, args.qtype, fout)

  else:
    parser.print_help()

  dbcon.close()



#!/usr/bin/env python3
"""
UMLS REST API client
UTS = UMLS Technology Services

 https://utslogin.nlm.nih.gov
 https://github.com/HHS/uts-rest-api
 https://documentation.uts.nlm.nih.gov/rest/home.html
 https://documentation.uts.nlm.nih.gov/rest/authentication.html
 https://documentation.uts.nlm.nih.gov/rest/concept/
 https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/
 https://documentation.uts.nlm.nih.gov/rest/search/
 https://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/abbreviations.html

 TGT = Ticket Granting Ticket
 (API requires one ticket per request.)
 CUI = Concept Unique Identifier
 Atom is a term in a source.
 Term-to-concept is many to one.

 Retrieves information for a known Semantic Type identifier (TUI)
 /semantic-network/{version}/TUI/{id}
 (DOES NOT SEARCH FOR INSTANCES OF THIS TYPE -- RETURNS METADATA ONLY.)
 Example TUIs:
 CHEM|Chemicals & Drugs|T103|Chemical
 CHEM|Chemicals & Drugs|T200|Clinical Drug
 CHEM|Chemicals & Drugs|T126|Enzyme
 CHEM|Chemicals & Drugs|T125|Hormone
 CHEM|Chemicals & Drugs|T129|Immunologic Factor
 CHEM|Chemicals & Drugs|T127|Vitamin
 DISO|Disorders|T020|Acquired Abnormality
 DISO|Disorders|T190|Anatomical Abnormality
 DISO|Disorders|T049|Cell or Molecular Dysfunction
 DISO|Disorders|T047|Disease or Syndrome
 DISO|Disorders|T048|Mental or Behavioral Dysfunction
 DISO|Disorders|T046|Pathologic Function
 DISO|Disorders|T184|Sign or Symptom
 GENE|Genes & Molecular Sequences|T087|Amino Acid Sequence
 GENE|Genes & Molecular Sequences|T088|Carbohydrate Sequence
 GENE|Genes & Molecular Sequences|T028|Gene or Genome
 GENE|Genes & Molecular Sequences|T085|Molecular Sequence
 GENE|Genes & Molecular Sequences|T086|Nucleotide Sequence
"""
###
import sys,os,argparse,re,yaml,json,csv,logging,requests,time
#
from .. import umls
#
#############################################################################
if __name__=='__main__':
  API_HOST='uts-ws.nlm.nih.gov'
  API_BASE_PATH="/rest"
  API_AUTH_SERVICE="http://umlsks.nlm.nih.gov"
  API_AUTH_HOST="utslogin.nlm.nih.gov"
  API_AUTH_ENDPOINT='/cas/v1/api-key'
  API_HEADERS={"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain", "User-Agent":"python" }
  SRCS_FILE='/home/data/UMLS/data/umls_sources.csv'

  parser = argparse.ArgumentParser(description='UMLS REST API client utility', epilog='CUI = Concept Unique Identifier')
  ops = ['getConcept', 'getAtoms', 'getRelations', 'search', 'searchByTUI', 
	'listSources',  'cui2Code']
  searchTypes = ['exact', 'words', 'leftTruncation', 'rightTruncation', 'approximate',
	'normalizedString']
  inputTypes = ['atom', 'code', 'sourceConcept', 'sourceDescriptor', 'sourceUi', 'tty']
  returnIdTypes = ['aui', 'concept', 'code', 'sourceConcept', 'sourceDescriptor', 'sourceUi']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--id", help="ID (ex:C0018787)")
  parser.add_argument("--idfile", help="input IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--idsrc", help="query ID source (default: CUI)")
  parser.add_argument("--searchType", choices=searchTypes, default='words')
  parser.add_argument("--inputType", choices=inputTypes, default='atom')
  parser.add_argument("--returnIdType", choices=returnIdTypes, default='concept')
  parser.add_argument("--srcs", help='list of sources to include in response')
  #parser.add_argument("--umls_srcs_file", help="list of all UMLS sources")
  parser.add_argument("--searchQuery", help='string or code')
  parser.add_argument("--skip", default=0, type=int)
  parser.add_argument("--nmax", default=0, type=int)
  parser.add_argument("--version", default="current", help="API version")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--api_auth_host", default=API_AUTH_HOST)
  parser.add_argument("--api_auth_endpoint", default=API_AUTH_ENDPOINT)
  parser.add_argument("--api_auth_service", default=API_AUTH_SERVICE)
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.umls.yaml")
  parser.add_argument("--api_key", help="API key")
  parser.add_argument("-v", "--verbose", default=0, action="count")

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  if args.ofile:
    fout = open(args.ofile, "w+")
    if not fout: parser.error('Cannot open: %s'%args.ofile)
  else:
    fout = sys.stdout

  params={};
  if os.path.exists(args.param_file):
    with open(args.param_file, 'r') as fh:
      for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
        for k,v in param.items():
          params[k] = v
  api_key = args.api_key if args.api_key else params['API_KEY'] if 'API_KEY' in params else ''
  if not api_key:
    parser.error('Please specify valid API_KEY via --api_key or --param_file') 

  api_auth_url = 'https://'+args.api_auth_host+args.api_auth_endpoint
  auth = umls.Utils.Authentication(api_key, args.api_auth_service, api_auth_url, API_HEADERS)
  auth.setVerbosity(args.verbose)

  ids=[];
  if args.idfile:
    fin = open(args.idfile)
    if not fin:
      parser.error('Cannot open: %s'%args.idfile)
    while True:
      line=fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('input IDs: %d'%(len(ids)))
    fin.close()
  elif args.id:
    ids.append(args.id)

  srclist = umls.Utils.SourceList()
  srclist.initFromApi(base_url, args.version, auth)

  if args.srcs:
    for src in re.split(r'[,\s]+', args.srcs.strip()):
      if not srclist.has_src(src):
          parser.error('Source unknown: "%s"'%src)

  if args.op == 'listSources':
    fout.write("abbreviation\tshortName\tpreferredName\n")
    for i,src in enumerate(srclist.sources):
      abbr,name,prefname = src
      fout.write('%s\t%s\t%s\n'%(abbr, name, prefname))
    logging.info('n_src: %d'%len(srclist.sources))

  elif args.op == 'getConcept':
    umls.Utils.GetConcept(base_url, args.version, args.idsrc, auth, ids, args.skip, args.nmax, fout)

  elif args.op == 'getRelations':
    if args.idsrc and args.idsrc!='CUI':
      parser.error('getRelations requires --idsrc CUI.')
    umls.Utils.GetRelations(base_url, args.version, auth, ids, args.skip, args.nmax, args.srcs, fout)

  elif args.op == 'getAtoms':
    if args.idsrc and args.idsrc!='CUI':
      parser.error('getAtoms requires --idsrc CUI.')
    umls.Utils.GetAtoms(base_url, args.version, auth, ids, args.skip, args.nmax, args.srcs, fout)

  elif args.op == 'cui2Code':
    if args.idsrc and args.idsrc!='CUI':
      parser.error('cui2Code requires --idsrc CUI.')
    i_cui=0; n_code=0;
    for cui in ids:
      i_cui+=1
      fout.write('%d.\t%s:'%(i_cui,cui))
      codes = umls.Utils.Cui2Code(base_url, args.version, auth, cui, args.srcs, fout)
      n_code_this=0;
      for src in sorted(codes.keys()):
        for i,atom in enumerate(sorted(list(codes[src]))):
          fout.write('%s:\t%d.\t%s\t%s'%(src, i+1, atom.code, atom.name))
          n_code_this+=1
      n_code+=n_code_this
    logging.info('n_cui: %d'%i_cui)
    logging.info('n_code: %d'%n_code)

  elif args.op == 'search':
    if not args.searchQuery:
      parser.error('search requires --searchQuery')
    umls.Utils.Search(base_url, args.version, auth, args.searchQuery, args.searchType, args.inputType, args.returnIdType, args.srcs, fout)

  elif args.op == 'searchByTUI':
    parser.error('ERROR: searchByTUI NOT IMPLEMENTED YET.')

  else:
    parser.error('Invalid operation: %s'%args.op)

